from typing import List, Optional

from .error import APIError
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, FeatureSet, DataSource
from pyrasgo.schemas.enums import Granularity, ModelType
from pyrasgo.utils import track_usage


class Update():

    def __init__(self):
        from .connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    @track_usage
    def collection_attributes(self,
                              id: int,
                              attributes: List[dict]):
        """
        Create or update attributes on a Rasgo Collection

        param attributes: dict [{"key": "value"}, {"key": "value"}]
        """
        msg = 'attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]'
        if not isinstance(attributes, list):
            raise APIError(msg)
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise APIError(msg)
            for k, v in kv.items():
                attr.append(api.Attribute(key=k, value=v))
        attr_in = api.CollectionAttributeBulkCreate(collectionId = id, attributes=attr)
        return self.api._put(f"/models/{id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()

    @track_usage
    def column(self,
               id: int,
               name: Optional[str] = None,
               data_type: Optional[str] = None,
               feature_set_id: Optional[int] = None,
               dimension_id: Optional[int] = None) -> api.Column:
        column = api.ColumnUpdate(id=id,
                                  name=name, dataType=data_type,
                                  featureSetId=feature_set_id,
                                  dimensionId=dimension_id)
        response = self.api._patch(f"/columns/{id}", column.dict(exclude_unset=True, exclude_none=True), api_version=1).json()
        return api.Column(**response)

    @track_usage
    def data_source(self,
                    id: int,
                    name: Optional[str] = None,
                    domain: Optional[str] = None,
                    source_type: Optional[str] = None,
                    table: Optional[str] = None,
                    database: Optional[str] = None,
                    schema: Optional[str] = None,
                    source_code: Optional[str] = None,
                    table_status: Optional[str] = None,
                    parent_source_id: Optional[int] = None):
        data_source = api.DataSourceUpdate(id=id,
                                           name=name,
                                           domain=domain,
                                           table=table,
                                           tableDatabase=database,
                                           tableSchema=schema,
                                           sourceCode=source_code,
                                           tableStatus=table_status,
                                           sourceType=source_type,
                                           parentId=parent_source_id)
        response = self.api._patch(f"/data-source/{id}", data_source.dict(exclude_unset=True, exclude_none=True), api_version=1).json()
        return DataSource(api_object=response)

    @track_usage
    def dataframe(self,
                  unique_id: str,
                  name: Optional[str] = None,
                  shared_status: str = None,
                  column_hash: str = None,
                  update_date: str = None) -> api.Dataframe:
        if shared_status not in [None, 'private', 'organization', 'public']:
            raise APIError("Valid values for shared_status are ['private', 'organization', 'public']")
        dataframe = api.DataframeUpdate(name=name,
                                        uniqueId=unique_id,
                                        sharedStatus=shared_status,
                                        columnHash=column_hash,
                                        updatedDate = update_date)
        response = self.api._patch(f"/dataframes/{unique_id}", dataframe.dict(exclude_unset=True, exclude_none=True), api_version=1).json()
        return api.Dataframe(**response)

    @track_usage
    def feature(self,
                id: int,
                display_name: Optional[str] = None,
                column_name: Optional[str] = None,
                description: Optional[str] = None,
                status: Optional[str] = None,
                tags: Optional[List[str]] = None,
                git_repo: Optional[str] = None) -> Feature:
        feature = api.FeatureUpdate(id=id,
                                    name=display_name,
                                    code=column_name,
                                    description=description,
                                    orchestrationStatus=status,
                                    tags=tags,
                                    gitRepo=git_repo)
        response = self.api._patch(f"/features/{id}", feature.dict(exclude_unset=True, exclude_none=True), api_version=1).json()
        return Feature(api_object=response)

    @track_usage
    def feature_attributes(self,
                           id: int,
                           attributes: List[dict]):
        """
        Create or update attributes on a feature

        param attributes: dict [{"key": "value"}, {"key": "value"}]
        """
        msg = 'attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]'
        if not isinstance(attributes, list):
            raise APIError(msg)
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise APIError(msg)
            for k, v in kv.items():
                attr.append(api.Attribute(key=k, value=v))
        attr_in = api.FeatureAttributeBulkCreate(featureId = id,
                                             attributes=attr)
        return self.api._put(f"/features/{id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()

    @track_usage
    def feature_set(self,
                    id: int,
                    name: Optional[str] = None,
                    data_source_id: Optional[int] = None,
                    table_name: Optional[str] = None,
                    file_path: Optional[str] = None,
                    source_code: Optional[str] = None) -> FeatureSet:
        feature_set = api.FeatureSetUpdate(id=id,
                                           name=name,
                                           snowflakeTable=table_name,
                                           dataSourceId=data_source_id,
                                           rawFilePath=file_path,
                                           sourceCode=source_code)
        response = self.api._patch(f"/feature-sets/{id}", feature_set.dict(exclude_unset=True, exclude_none=True), api_version=0).json()
        return FeatureSet(api_object=response)


import pandas as pd
from typing import List, Optional

from .error import APIError
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, FeatureSet, DataSource
from pyrasgo.utils import track_usage


class Get():
    
    def __init__(self):
        from .connection import Connection
        from pyrasgo.config import get_session_api_key
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()

    @track_usage
    def collection(self, id: int) -> Collection:
        """
        Returns a Rasgo Collection (set of joined Features) matching the specified id
        """
        try:
            return Collection(api_object=self.api._get(f"/models/{id}", api_version=1).json())
        except:
            raise APIError(f"Collection {id} does not exist or this API key does not have access.")

    @track_usage
    def collection_attributes(self, id: int) -> api.CollectionAttributes:
        """
        Returns a dict of attributes for a collection
        """
        try:
            response = self.api._get(f"/models/{id}/attributes", api_version=1).json()
            dict_out = {}
            for kv in response:
                dict_out[kv['key']] = kv['value']
            return api.CollectionAttributes(collectionId=id, attributes=dict_out)
        except:
            raise APIError(f"Collection {id} does not exist or this API key does not have access.")

    @track_usage
    def collections(self, include_shared: bool=False) -> List[Collection]:
        """
        Returns all Rasgo Collections (set of joined Features) that I have author access to. Add an include_shared 
        parameter to return all Rasgo Collections that I have any access to (author or shared access)
        :param include_shared: Boolean value indicating if the return should include all accessible collections
        """
        try:
            return [Collection(api_object=entry) for entry in self.api._get(f"/models", {"include_shared": include_shared}, api_version=1).json()]
        except:
            raise APIError("Collections do not exist or this API key does not have access.")

    @track_usage
    def collections_by_attribute(self, key: str, value: str = None) -> List[Collection]:
        """
        Returns a list of Rasgo Collections that match an attribute
        """
        try:
            params = {"key": key}
            if value:
                params.update({"value": value})
            return self.api._get(f"/models/attributes/models", params=params, api_version=1).json()
        except:
            raise APIError(f"Key {key}: {value or 'Any'} does not exist or this API key does not have access.")

    @track_usage
    def column(self, id: int) -> api.Column:
        """
        Returns a Column with the specified id
        """
        try:
            response = self.api._get(f"/columns/{id}", api_version=1).json()
            return api.Column(**response)
        except:
            raise APIError(f"Column {id} does not exist or this API key does not have access.")

    @track_usage
    def columns_by_featureset(self, id: int) -> List[api.Column]:
        """
        Returns all Columns in the specified FeatureSet
        """
        try:
            response = self.api._get(f"/columns/by-featureset/{id}", api_version=1).json()
            return [api.Column(**entry) for entry in response]
        except:
            raise APIError(f"FeatureSet {id} does not exist or this API key does not have access.")

    @track_usage
    def data_sources(self) -> List[DataSource]:
        """
        Returns all DataSources available in your organization or Rasgo Community
        """
        try:
            response = self.api._get("/data-source", api_version=1).json()
            return [DataSource(api_object=entry) for entry in response]
        except:
            raise APIError("Data Sources do not exist or this API key does not have access.")

    @track_usage
    def data_source(self, id: int) -> DataSource:
        """
        Returns the DataSource with the specified id
        """
        try:
            response = self.api._get(f"/data-source/{id}", api_version=1).json()
            return DataSource(api_object=response)
        except:
            raise APIError(f"Data Source {id} does not exist or this API key does not have access.")

    @track_usage
    def data_source_columns(self, id: int) -> List[api.DataSourceColumn]:
        """
        Returns columns in the DataSource with the specified id
        """
        try:
            response = self.api._get(f"/data-source/columns/{id}", api_version=1).json()
            return [api.DataSourceColumn(**column) for column in response['columns']]
        except:
            raise APIError(f"Data Source {id} does not exist or this API key does not have access.")

    @track_usage
    def data_source_stats(self, id: int):
        """
        Returns the stats profile of the specificed data source
        """
        try:
            return self.api._get(f"/data-source/profile/{id}", api_version=1).json()
        except:
            raise APIError(f"Stats do not exist for DataSource {id}")

    @track_usage
    def dataframes(self) -> List[api.Dataframe]:
        """
        Returns all Dataframes available in your organization or Rasgo Community
        """
        try:
            response = self.api._get("/dataframes", api_version=1).json()
            return [api.Dataframe(**entry) for entry in response]
        except:
            raise APIError("Dataframes do not exist or this API key does not have access.")

    @track_usage
    def dataframe(self, unique_id: str) -> api.Dataframe:
        """
        Returns the Dataframe with the specified id
        """
        try:
            response = self.api._get(f"/dataframes/{unique_id}", api_version=1).json()
            return api.Dataframe(**response)
        except:
            raise APIError(f"Dataframe {unique_id} does not exist or this API key does not have access.")

    @track_usage
    def dimensionalities(self) -> List[api.Dimensionality]:
        """
        Returns all Dimensionalities available in your organization or Rasgo Community
        """
        try:
            response = self.api._get("/dimensionalities", api_version=1).json()
            return [api.Dimensionality(**entry) for entry in response]
        except:
            raise APIError("Dimensionalities do not exist or this API key does not have access.")
    
    @track_usage
    def feature(self, id: int) -> Feature:
        """
        Returns the Feature with the specified id
        """
        try:
            return Feature(api_object=self.api._get(f"/features/{id}", api_version=1).json())
        except:
            raise APIError(f"Feature {id} does not exist or this API key does not have access.")
    
    @track_usage
    def feature_attributes(self, id: int) -> api.FeatureAttributes:
        """
        Returns a dict of attributes for a feature
        """
        try:
            response = self.api._get(f"/features/{id}/attributes", api_version=1).json()
            dict_out = {}
            for kv in response:
                dict_out[kv['key']] = kv['value']
            return api.FeatureAttributes(featureId=id, attributes=dict_out)
        except:
            raise APIError(f"Feature {id} does not exist or this API key does not have access.")

    @track_usage
    def feature_attributes_log(self, id: int) -> tuple:
        """
        Returns a list of all attributes values logged to a feature over time
        """
        try:
            response = self.api._get(f"/features/{id}/attributes/log", api_version=1).json()
            lst_out = []
            for kv in response:
                dict_item={}
                dict_item[kv['key']] = kv.get('value', None)
                dict_item['updatedBy'] = kv.get('recordAuthorId', None)
                dict_item['updated'] = kv.get('recordTimestamp', None)
                lst_out.append(dict_item)
            return api.FeatureAttributesLog(featureId=id, attributes=lst_out)
        except:
            raise APIError(f"Feature {id} does not exist or this API key does not have access.")

    @track_usage
    def feature_set(self, id: int) -> FeatureSet:
        """
        Returns the FeatureSet (set of Fetures) with the specified id
        """
        try:
            response = self.api._get(f"/feature-sets/{id}", api_version=1).json()
            return FeatureSet(api_object=response)
        except:
            raise APIError(f"FeatureSet {id} does not exist or this API key does not have access.")

    @track_usage
    def feature_set_yml(self, id: int) -> FeatureSet:
        """
        Returns the FeatureSet (set of Fetures) with the specified id
        """
        try:
            response = self.api._get(f"/feature-sets/{id}", api_version=1).json()
            return api.FeatureSetYML(**response)
        except:
            raise APIError(f"FeatureSet {id} does not exist or this API key does not have access.")

    @track_usage
    def feature_sets(self) -> List[FeatureSet]:
        """
        Returns a list of FeatureSets (set of Features) available in your organization or Rasgo Community
        """
        try:
            response = self.api._get("/feature-sets", api_version=1).json()
            return [FeatureSet(api_object=entry) for entry in response]
        except:
            raise APIError("FeatureSets do not exist or this API key does not have access.")

    @track_usage
    def feature_stats(self, id: int) -> Optional[api.FeatureStats]:
        """
        Returns the stats profile for the specified Feature
        """
        try:
            stats_json = self.api._get(f"/features/{id}/stats", api_version=1).json()
            return api.FeatureStats(**stats_json["featureStats"])
        except:
            raise APIError(f"Stats do not exist yet for feature {id}.")
    
    @track_usage
    def features(self) -> FeatureList:
        """
        Returns a list of Features available in your organization or Rasgo Community
        """
        try:
            return FeatureList(api_object=self.api._get("/features", api_version=1).json())
        except:
            raise APIError("Features do not exist or this API key does not have access.")

    @track_usage
    def features_by_attribute(self, key: str, value: str = None) -> List[Feature]:
        """
        Returns a list of features that match an attribute
        """
        try:
            params = {"key": key}
            if value:
                params.update({"value": value})
            return FeatureList(api_object=self.api._get(f"/features/attributes/features", params=params, api_version=1).json())
        except:
            raise APIError(f"Key {key}: {value or 'Any'} does not exist or this API key does not have access.")

    @track_usage
    def features_by_featureset(self, id: int) -> FeatureList:
        """
        Returns a list of Features in the specific FeatureSet
        """
        try:
            response = self.api._get(f"/features/by-featureset/{id}", api_version=1)
            return FeatureList(api_object=response.json())
        except:
            raise APIError(f"FeatureSet {id} does not exist or this API key does not have access.")

    @track_usage
    def shared_collections(self) -> List[Collection]:
        """
        Returns all Rasgo Collections (set of joined Features) shared in my organization or in Rasgo community
        """
        try:
            return [Collection(api_object=entry) for entry in self.api._get(f"/models/shared", api_version=1).json()]
        except:
            raise APIError("Shared Collections do not exist or this API key does not have access.")

    @track_usage
    def source_columns(self, table: Optional[str] = None, database: Optional[str] = None, schema: Optional[str] = None, data_type: Optional[str] = None) -> pd.DataFrame:
        """
        Returns a DataFrame of columns in Snowflake tables and views that are queryable as feature sources
        """
        return self.data_warehouse.get_source_columns(table=table, database=database, schema=schema, data_type=data_type)

    @track_usage
    def source_tables(self, database: Optional[str] = None, schema: Optional[str] = None) -> pd.DataFrame:
        """
        Return a DataFrame of Snowflake tables and views that are queryable as feature sources
        """
        return self.data_warehouse.get_source_tables(database=database, schema=schema)

    @track_usage
    def transform_list(self) -> dict:
        """
        Returns json of all available transforms
        """
        return self.api._get("/features/transforms2", api_version=1).json()

    @track_usage
    def user(self):
        return self.api._get("/users/me", api_version=1).json()


from enum import Enum
from typing import Union


class DataType(Enum):
    SMALLINT = 'smallint'
    BIGINT = 'bigint'
    INT = 'int'
    INTEGER = 'integer'
    TINYINT = 'tinyint'
    BYTEINT = 'byteint'
    FLOAT = 'float'
    FLOAT4 = 'float4'
    FLOAT8 = 'float8'
    FLOAT64 = 'float64'
    DECIMAL = 'decimal'
    NUMERIC = 'numeric'
    NUMBER = 'number'
    REAL = 'real'
    DOUBLE = 'double'
    STRING = 'string'
    TEXT = 'text'
    VARCHAR = 'varchar'
    CHAR = 'char'
    CHARACTER = 'character'
    DATE = 'date'
    DATETIME = 'datetime'
    TIME = 'time'
    TIMESTAMP = 'timestamp'
    TIMESTAMP_LTZ = 'timestamp_ltz'
    TIMESTAMP_NTZ = 'timestamp_ntz'
    TIMESTAMP_TZ = 'timestamp_tz'
    BINARY = 'binary'
    VARBINARY = 'varbinary'
    BOOLEAN = 'boolean'
    BOOL = 'bool'
    VARIANT = 'variant'
    OBJECT = 'object'
    ARRAY = 'array'

class Granularity(Enum):
    """
    Enum class to describe allowed types of granularities
    """
    # TODO: This exists to support validation prior to posting from the API.
    #       For ease of maintenance, this should be handled through API returns.
    ACTION = 'action'
    ACTOR = 'actor'
    AIRLINE = 'airline'
    AIRPORT = 'airport'
    ATBAT = 'atbat'
    AUTHOR = 'author'
    BUYER = 'buyer'
    CBG = 'CBG'
    CITY = 'city'
    CLAIM = 'claim'
    COMPANY = 'company'
    COUNTY = 'county'
    COUNTRY = 'country'
    CUSTOM = 'custom'
    CUSTOMER = 'customer'
    DAY = 'day'
    DMA = 'DMA'
    FIPS = 'FIPS'
    FLIGHT = 'flight'
    GAME = 'game'
    HOUR = 'hour'
    LATLONG = 'latlong'
    MAKE = 'make'
    MEMBER = 'member'
    MINUTE = 'minute'
    MODEL = 'model'
    MONTH = 'month'
    MOVIE = 'movie'
    NAME = 'name'
    PAYER = 'payer'
    PLAY = 'play'
    PLAYER = 'player'
    PITCH = 'pitch'
    PRODUCT = 'product'
    PROVIDER = 'provider'
    PUBLISHER = 'publisher'
    QUARTER = 'quarter'
    SEASON = 'season'
    SECOND = 'second'
    STATE = 'state'
    STORE = 'store'
    STUDIO = 'studio'
    SUBSCRIBER = 'subscriber'
    TEAM = 'team'
    TERM = 'term'
    TICKER = 'ticker'
    TITLE = 'title'
    USER = 'user'
    UNIT = 'unit'
    WEEK = 'week'
    YEAR = 'year'
    ZIPCODE = 'zipcode'


class ModelTypes(Enum):
    """
    Enum class to describe allowed types of models
    """
    # TODO: This exists to support validation prior to posting from the API.
    #       For ease of maintenance, this should be handled through API returns.
    TIMESERIES = 'Timeseries'

    @classmethod
    def _missing_(cls, value):
        """ Sets up case-insensitive lookup """
        formatted = value.replace(" ", "").lower().capitalize()
        try:
            return cls._value2member_map_[formatted]
        except KeyError:
            raise ValueError("%r is not a valid %s" % (value, cls.__name__))


class ModelType(object):
    """
    Class to wrap the ModelTypes enum for additional functionality

    (Extending the enum class is not allowed, and subclassing is possible, but unnecessary)
    """

    def __init__(self, name: str):
        self._type = ModelTypes(name)
        # TODO: Lookup compatible when other model types and/or granularities are supported.
        self._compatible_granularities = [Granularity.DAY, Granularity.MONTH,
                                          Granularity.YEAR, Granularity.WEEK, Granularity.QUARTER]

    @property
    def name(self):
        return self._type.name

    @property
    def value(self):
        return self._type.value

    @property
    def compatible_granularities(self):
        return self._compatible_granularities

    def is_compatible(self, granularity: Union[str, Granularity]) -> bool:
        # If not enum, convert to enum first.
        try:
            granularity.name
        except AttributeError:
            granularity = Granularity(granularity)
        return granularity in self._compatible_granularities



from pyrasgo.api.admin import Admin
from pyrasgo.api.create import Create
from pyrasgo.api.delete import Delete
from pyrasgo.api.get import Get
from pyrasgo.api.match import Match
from pyrasgo.api.publish import Publish
from pyrasgo.api.read import Read
from pyrasgo.api.register import Register
from pyrasgo.api.save import Save
from pyrasgo.api.update import Update


from typing import List

import pandas as pd

from pyrasgo.storage.dataframe._base_transform import BaseTransform


class Transform:
    def rolling_average(self, df: pd.DataFrame, window_size: int, columns: List[str] = None, aliases: List[str] = None):
        """Calculate mean over a window size in by number of rows"""
        if aliases:
            for column, alias in zip(columns, aliases):
                BaseTransform.rolling_average(df, window_size, column, alias)
        else:
            for column in columns:
                BaseTransform.rolling_average(df, window_size, column)
        return df

    def cumulative_average(self, df: pd.DataFrame, columns: List[str] = None, aliases: List[str] = None):
        """Calculate mean over window from beginning of df to current location"""
        if aliases:
            for column, alias in zip(columns, aliases):
                BaseTransform.cumulative_average(df, column, alias)
        else:
            for column in columns:
                BaseTransform.cumulative_average(df, column)
        return df

    def add(self, df: pd.DataFrame, value: float, columns: List[str] = None, aliases: List[str] = None):
        """Add a number to each value of a dataframe"""
        if aliases:
            for column, alias in zip(columns, aliases):
                BaseTransform.add(df, value, column, alias)
        else:
            for column in columns:
                BaseTransform.add(df, value, column)
        return df

    def subtract(self, df: pd.DataFrame, value: float, columns: List[str] = None, aliases: List[str] = None):
        """Subtract a number from each element of a dataframe"""
        if aliases:
            for column, alias in zip(columns, aliases):
                BaseTransform.subtract(df, value, column, alias)
        else:
            for column in columns:
                BaseTransform.subtract(df, value, column)
        return df

    def divide(self, df: pd.DataFrame, value: float, columns: List[str] = None, aliases: List[str] = None):
        """Divide each element in a dataframe by a value"""
        if aliases:
            for column, alias in zip(columns, aliases):
                BaseTransform.divide(df, value, column, alias)
        else:
            for column in columns:
                BaseTransform.divide(df, value, column)
        return df

    def multiply(self, df: pd.DataFrame, value: float, columns: List[str] = None, aliases: List[str] = None):
        """Multiply dataframe by a value"""
        if aliases:
            for column, alias in zip(columns, aliases):
                BaseTransform.multiply(df, value, column, alias)
        else:
            for column in columns:
                BaseTransform.multiply(df, value, column)
        return df

    def mult(self, df: pd.DataFrame, value: float, columns: List[str] = None, aliases: List[str] = None):
        """Alias for multiply"""
        return self.multiply(df, value)



from typing import List, Optional

from .error import APIError
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, FeatureSet, DataSource
from pyrasgo.schemas.enums import Granularity, ModelType
from pyrasgo.utils import track_usage


class Match():

    def __init__(self):
        from .connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    @track_usage
    def data_source(self, fqtn: str) -> DataSource:
        """
        Returns the first Data Source that matches the specified name
        """
        try:
            response = self.api._get(f"/data-source", {"table": fqtn}, api_version=1).json()
            return DataSource(api_object=response[0])
        except:
            return None

    @track_usage
    def dataframe(self, name: str = None, unique_id: str = None) -> api.Dataframe:
        """
        Returns the first Dataframe that matches the specified name or uid
        """
        try:
            if name:
                response = self.api._get(f"/dataframes", {"name": name}, api_version=1).json()
                return api.Dataframe(**response[0])
            elif unique_id:
                response = self.api._get(f"/dataframes/{unique_id}", api_version=1).json()
                return api.Dataframe(**response)
        except:
            return None
        return None

    @track_usage
    def dimensionality(self, granularity: str) -> api.Dimensionality:
        """
        Returns the first community or private Dimensionality that matches the specified granularity
        """
        try:
            response = self.api._get(f"/dimensionalities/granularity/{granularity}", api_version=1).json()
            return api.Dimensionality(**response)
        except:
            return None

    @track_usage
    def column(self, name: str, feature_set_id: int) -> Optional[api.Column]:
        """
        Returns the first Column matching the specidied name in the specified featureset
        """
        try:
            cols = self.api._get(f"/columns/by-featureset/{feature_set_id}", api_version=1).json()
            for c in cols:
                if name == c["name"]:
                    return api.Column(**c)
            return None
        except:
            return None

    @track_usage
    def feature(self, code: str, feature_set_id: int) -> Optional[Feature]:
        """
        Returns the first Feature matching the specified name in the specified featureset
        """
        try:
            features = self.api._get(f"/features/by-featureset/{feature_set_id}", api_version=1).json()
            for f in features:
                if code == f["code"]:
                    return Feature(api_object=f)
            return None
        except:
            return None

    @track_usage
    def feature_set(self, table_name: Optional[str] = None, source_id: int = None) -> Optional[FeatureSet]:
        """
        Returns the first FeatureSet matching the specified table name
        """
        try:
            if table_name:
                fs = self.api._get(f"/feature-sets/", {"source_table": table_name}, api_version=1).json()
            elif source_id:
                fs = self.api._get(f"/feature-sets/", {"source_id": source_id}, api_version=1).json()
            return FeatureSet(api_object=fs[0])
        except:
            return None


import logging
import os
import pandas as pd
from typing import Optional

from snowflake import connector as snowflake

from pyrasgo.api.session import Session
from pyrasgo.storage.dataframe import utils as dfutils
from pyrasgo.storage.datawarehouse.base import DataWarehouse, DataWarehouseSession
from pyrasgo.utils import naming
from pyrasgo.utils import track_usage

class SnowflakeDataWarehouse(DataWarehouse, Session, metaclass=DataWarehouseSession):
    def __init__(self):
        username = self.profile.get("dwUsername", self.profile.get("snowUsername"))
        if username is None:
            raise EnvironmentError("Your user is missing credentials, please contact Rasgo support.")
        self.username = username
        self.password = self.profile.get("dwPassword", self.profile.get("snowPassword"))

        organization = self.profile.get("organization")
        self.account = organization.get("account")
        self.database = organization.get("database")
        self.schema = organization.get("schema")
        self.warehouse = organization.get("warehouse")
        self.org_prefix = organization.get("rolePrefix", "code")
        if self.org_prefix is None:
            raise EnvironmentError("Your organization is missing credentials, please contact Rasgo support.")
        self.user_role = self.profile.get("snowRole", f"{self.org_prefix}_{self.username}")

    @property
    def publisher_role(self) -> str:
        return self.org_prefix+'PUBLISHER'

    @property
    def reader_role(self) -> str:
        return self.org_prefix+'READER'

    @property
    @track_usage
    def user_connection(self) -> snowflake.SnowflakeConnection:
        return snowflake.connect(**self.user_credentials)

    @property
    @track_usage
    def user_credentials(self) -> dict:
        return {
            "user": self.username,
            "password": self.password,
            "account": self.account,
            "database": self.database,
            "schema": self.schema,
            "warehouse": self.warehouse,
            "role": self.user_role
        }

    @track_usage
    def execute_query(self, query: str, params: Optional[dict] = None):
        """
        Execute a query on the [cloud] data platform.

        :param query: String to be executed on the data platform
        :param params: Optional parameters
        :return:
        """
        return self.user_connection.cursor().execute(query, params)

    @track_usage
    def query_into_dataframe(self, query: str, params: Optional[dict] = None) -> pd.DataFrame:
        """
        Execute a query on the [cloud] data platform.

        :param query: String to be executed on the data platform
        :param params: Optional parameters
        :return:
        """
        cur = self.user_connection.cursor()
        cur.execute(query, params)
        return cur.fetch_pandas_all()

    @track_usage
    def get_source_table(self, table_name: str, database: Optional[str] = None, schema: Optional[str] = None,
                         record_limit: Optional[int] = None) -> pd.DataFrame:
        if record_limit is None:
            logging.info(f"Loading all rows from {table_name}...")
        database = database or self.database
        schema = schema or self.schema
        result_set = self.execute_query(f"SELECT * FROM {database}.{schema}.{naming._quote_restricted_keywords(table_name)} {f'LIMIT {record_limit}' if record_limit else ''}")
        return pd.DataFrame.from_records(iter(result_set), columns=[x[0] for x in result_set.description])

    @track_usage
    def get_source_tables(self, database: Optional[str] = None, schema: Optional[str] = None):
        database = database or self.database
        schema = schema or self.schema
        filters = f"WHERE TABLE_SCHEMA='{schema}'"
        result_set = self.execute_query(f"SELECT TABLE_NAME, TABLE_CATALOG||'.'||TABLE_SCHEMA||'.'||TABLE_NAME AS FQTN, ROW_COUNT, TABLE_OWNER, CREATED, LAST_ALTERED FROM {database}.INFORMATION_SCHEMA.TABLES {filters}")
        return pd.DataFrame.from_records(iter(result_set), columns=[x[0] for x in result_set.description])

    @track_usage
    def get_source_columns(self, table: Optional[str] = None, database: Optional[str] = None, schema: Optional[str] = None, data_type: Optional[str] = None):
        database = database or self.database
        schema = schema or self.schema
        filters = f"WHERE TABLE_SCHEMA='{schema}'"
        filters += f" AND TABLE_NAME='{table}'" if table else ""
        filters += f" AND DATA_TYPE='{data_type}'" if data_type else ""
        result_set = self.execute_query(f"SELECT COLUMN_NAME, DATA_TYPE, TABLE_NAME, TABLE_CATALOG||'.'||TABLE_SCHEMA||'.'||TABLE_NAME AS FQTN FROM {database}.INFORMATION_SCHEMA.COLUMNS {filters}")
        return pd.DataFrame.from_records(iter(result_set), columns=[x[0] for x in result_set.description])

    @track_usage
    def grant_table_ownership(self, table: str, role: str, database: Optional[str] = None, schema: Optional[str] = None):
        """
        Grants ownership of a table or FQTN to specified role
        Note: If user role is not the table owner, this will fail
        """
        database = database or self.database
        schema = schema or self.schema
        self.execute_query(f"GRANT OWNERSHIP ON {database}.{schema}.{naming._quote_restricted_keywords(table)} TO ROLE {role} REVOKE CURRENT GRANTS;")

    @track_usage
    def grant_table_access(self, table: str, role: str, database: Optional[str] = None, schema: Optional[str] = None):
        """
        Grants select privilege on a table or FQTN to specified role
        Note: If user role does not have access with grant option, this will fail
        """
        database = database or self.database
        schema = schema or self.schema
        self.execute_query(f"GRANT SELECT ON {database}.{schema}.{naming._quote_restricted_keywords(table)} TO ROLE {role};")

    @track_usage
    def clone_table(self, existing_table: str, new_table: str, overwrite: bool = False):
        """
        Clones a table or FQTN
        Note: if user role doesn't have select & create access to db + schema + table this will fail
        """
        if overwrite:
            self.execute_query(f"CREATE OR REPLACE TABLE {naming._quote_restricted_keywords(new_table)} CLONE {naming._quote_restricted_keywords(existing_table)};")
        else:
            self.execute_query(f"CREATE TABLE {naming._quote_restricted_keywords(new_table)} CLONE {naming._quote_restricted_keywords(existing_table)};")

    @track_usage
    def append_to_table(self, from_table: str, into_table: str):
        """
        Inserts data from a table or FTQN into another table or FQTN
        Note: if user role doesn't have select & modify access to db + schema + table this will fail
        """
        self.execute_query(f"INSERT INTO {naming._quote_restricted_keywords(into_table)} SELECT * FROM {naming._quote_restricted_keywords(from_table)};")

    def _make_table_metadata(self, table: str):
        if table.count(".") > 0:
            table = table.split(".")[-1]
        metadata = {
            "database": self.database,
            "schema": self.schema,
            "table": table,
        }
        return metadata

    def _write_dataframe_to_table(self, df: pd.DataFrame, table_name: str, append: bool = False):
        """
        Note: we will allow users to select from other database, but only to write into org default database
        """
        from snowflake.connector.pandas_tools import write_pandas

        dfutils._cleanse_sql_dataframe(df)
        write_pandas(conn=self.user_connection, df=df, table_name=naming._cleanse_sql_name(table_name), quote_identifiers=False)


import pandas as pd


class BaseTransform:
    @classmethod
    def rolling_average(cls, df: pd.DataFrame, window_size, column: str, alias: str = None):
        new_col_name = alias or f"{column}_roll_avg"
        df[new_col_name] = df[column].rolling(window=window_size).mean()
        return df

    @classmethod
    def cumulative_average(cls, df: pd.DataFrame, column: str, alias: str = None):
        new_col_name = alias or f"{column}_cum_avg"
        df[new_col_name] = df[column].expanding().mean()
        return df

    @classmethod
    def add(cls, df: pd.DataFrame, value, column: str, alias: str = None):
        """Perform addition to each value in the dataframe"""
        new_col_name = alias or f"{column}_add_{value}"
        df[new_col_name] = df[column] + value
        return df

    @classmethod
    def subtract(cls, df: pd.DataFrame, value, column: str, alias: str = None):
        new_col_name = alias or f"{column}_subtract_{value}"
        df[new_col_name] = df[column] - value
        return df


    @classmethod
    def divide(cls, df: pd.DataFrame, value: float, column, alias: str = None):
        """Divide each element in a dataframe by a value"""
        new_col_name = alias or f"{column}_divide_{value}"
        df[new_col_name] = df[column] / value
        return df

    @classmethod
    def multiply(cls, df: pd.DataFrame, value: float, column, alias: str = None):
        """Multiply dataframe by a value"""
        new_col_name = alias or f"{column}_mult_{value}"
        df[new_col_name] = df[column] * value
        return df



from pyrasgo.primitives.collection import Collection
from pyrasgo.primitives.feature import Feature, FeatureList
from pyrasgo.primitives.feature_set import FeatureSet
from pyrasgo.primitives.source import DataSource


from typing import Dict, List, Optional

import pandas as pd

from pyrasgo.api.connection import Connection
from pyrasgo.schemas import feature as schema
from pyrasgo.schemas.attributes import Attribute, FeatureAttributes, FeatureAttributeBulkCreate
from pyrasgo.schemas.enums import Granularity
from pyrasgo.utils import track_usage

class FeatureList():
    """
        Convenience class to enable simpler console presentation,
        iteration and searching through lists of Features objects
    """

    def __init__(self, api_object):
        self.data = [Feature(api_object=entry) for entry in api_object]

    def __getitem__(self, i: int):
        return self.data[i]

    def __len__(self) -> int:
        return len(self.data)

    def __str__(self):
        ids = [str(feature.id) for feature in self]
        return (f"Features({len(self.data)} total, "
                f"ids: [{','.join(ids if len(self) < 7 else ids[0:3] + ['...'] + ids[-3:])}])")

    def __add__(self, other):
        if isinstance(other, Feature):
            return type(self)([feature.dict() for feature in self.data + [other]])
        if isinstance(other, type(self)):
            return type(self)([feature.dict() for feature in self.data + other.data])
        if isinstance(other, list) and all([isinstance(entry, Feature) for entry in other]):
            return type(self)([feature.dict() for feature in self.data + other])
        raise TypeError(f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'")

    def __repr__(self):
        return str(self)

    def filter(self, **kwargs):
        return [feature for feature in self.data
                if [feature.__getattr__(key) for key in kwargs.keys()] == list(kwargs.values())]


class Feature(Connection):
    """
    Stores a Rasgo Feature
    """

    def __init__(self, api_object, **kwargs):
        super().__init__(**kwargs)
        self._fields = schema.Feature(**api_object)

    def __getattr__(self, item):
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Feature(id={self.id}, name={self.name}, dataType={self.dataType}, description={self.description}, status={self.status})"

# ----------
# Properties
# ----------
    @property
    def attributes(self) -> dict:
        """
        Helper property to convert a list of dicts into a single dict
        """
        attr_dict = {}
        for a in self._fields.attributes:
            attr_dict.update({a.key: a.value})
        return attr_dict

    @property
    def columnName(self):
        """
        Retrieves the column name in the feature set table
        """
        return self._fields.code

    @property
    def dimensions(self):
        """
        Retrieves the dimensions of this feature set
        """
        columns = self._get(f"/columns/by-featureset/{self._fields.featureSet.id}", api_version=1).json()
        return [c for c in columns]
        # Need to call above endpoint until FS response contains dimensions
        #feature_set = self._fields.featureSet
        #if feature_set:
        #    return feature_set.dimensions
        #return None

    @property
    def displayName(self):
        """
        Alias for feature name
        """
        return self._fields.name

    @property
    def featureSet(self):
        """
        Retrieves the feature set for the feature
        """
        feature_set = self._fields.featureSet
        if feature_set:
            return feature_set.name
        return None

    @property
    def granularities(self):
        """
        Retrieves granularities for the feature
        """
        granularities = self._fields.granularities
        if granularities:
            return [Granularity(granularity.name) for granularity in granularities]
        return []

    @property
    def indexFields(self):
        """
        Retrieves the dimensions column name in the feature set table
        """
        return [d['name'] for d in self.dimensions]

    @property
    def sourceTable(self):
        """
        Retrieves the feature set table for the feature
        """
        feature_set = self._fields.featureSet
        if feature_set:
            return feature_set.snowflakeTable
        return None

    @property
    def status(self):
        """
        Returns feature status: Sandbox or Production
        """
        return 'Production' if self.orchestrationStatus == 'Productionized' else 'Sandbox'

    @property
    def tags(self) -> List[str]:
        return self._fields.tags

# -------
# Methods
# -------
    @track_usage
    def add_attributes(self, attributes: List[dict]):
        if not isinstance(attributes, list):
            raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]')
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]')
            for k, v in kv.items():
                attr.append(Attribute(key=k, value=v))
        attr_in = FeatureAttributeBulkCreate(featureId = self.id, attributes=attr)
        self._put(f"/features/{self.id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()
        self.refresh()

    @track_usage
    def add_tag(self, tag: str):
        if tag in self._fields.tags:
            return
        self._fields = schema.Feature(**self._put(f"/features/{self.id}/tags",
                                                api_version=1, _json={"tags": [tag, *self._fields.tags]}).json())

    @track_usage
    def add_tags(self, tags: List[str]):
        if all(tag in self._fields.tags for tag in tags):
            return
        self._fields = schema.Feature(**self._put(f"/features/{self.id}/tags",
                                                api_version=1, _json={"tags": [*tags, *self._fields.tags]}).json())

    @track_usage
    def build_stats(self):
        return self._post(f"/features/{self.id}/stats", api_version=1).json()

    @track_usage
    def display_source_code(self):
        """
        Convenience function to display the sourceCode property
        """
        feature_set = self._fields.featureSet
        if feature_set:
            return feature_set.sourceCode
        return None

    @track_usage
    def delete_tag(self, tag: str):
        if tag not in self._fields.tags:
            return
        self._fields = schema.Feature(**self._delete(f"/features/{self.id}/tags",
                                                   api_version=1, _json={"tags": [tag]}).json())

    @track_usage
    def delete_tags(self, tags: List[str]):
        if all(tag not in self._fields.tags for tag in tags):
            return
        self._fields = schema.Feature(**self._delete(f"/features/{self.id}/tags",
                                                   api_version=1, _json={"tags": [*tags]}).json())

    @track_usage
    def get_stats(self):
        try:
            stats_json = self._get(f"/features/{self.id}/stats", api_version=1).json()
        except:
            return 'Cannot find stats for this feature'
        stats_obj = schema.FeatureStats(**stats_json['featureStats']) or None
        return stats_obj

    @track_usage
    def productionize(self):
        """
        Sets a Feature's status from Sandbox to Production
        """
        print(f"Setting feature from {self.status} to Production")
        feature = schema.FeatureUpdate(id=self.id, orchestrationStatus='Productionized')
        self._fields = schema.Feature(**self._patch(f"/features/{self.id}",
                                                 api_version=1, _json=feature.dict(exclude_unset=True, exclude_none=True)).json())

    @track_usage
    def read_into_df(self,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        """
        Pull Feature data from DataWarehouse into a pandas Dataframe
        """
        from pyrasgo.api.read import Read
        return Read().feature_data(id=self.id, filters=filters, limit=limit)

    @track_usage
    def refresh(self):
        """
        Updates the Feature's attributes from the API
        """
        self._fields = schema.Feature(**self._get(f"/features/{self.id}", api_version=1).json())

    @track_usage
    def rename(self, new_name: str):
        """
        Updates a Feature's display name
        """
        print(f"Renaming Feature {self.id} from {self.displayName} to {new_name}")
        feature = schema.FeatureUpdate(id=self.id, name=new_name)
        self._fields = schema.Feature(**self._patch(f"/features/{self.id}",
                                                 api_version=1, _json=feature.dict(exclude_unset=True, exclude_none=True)).json())

    @track_usage
    def to_dict(self):
        feature_as_dict = {}
        feature_as_dict["sourceTable"] = self.sourceTable
        feature_as_dict["features"] = [{"columnName": self.columnName, "dataType": self.dataType, "displayName": self.name, "description": self.description}]
        feature_as_dict["dimensions"] = []
        for d in self.dimensions:
            feature_as_dict["dimensions"].append({"columnName": d.get("name"), "dataType": d.get("dataType"), "granularity": d.get("granularity")})
        feature_as_dict["status"] = self.orchestrationStatus
        feature_as_dict["tags"] = self.tags
        feature_as_dict["attributes"] = []
        for a in self._fields.attributes:
            feature_as_dict["attributes"].append({a.key: a.value})
        feature_as_dict["gitRepo"] = self.gitRepo
        return feature_as_dict

    def _make_table_metadata(self):
        table_attribute = self._fields.featureSet.snowflakeTable
        organization = self._get_profile().get("organization")
        if table_attribute.count(".") == 2:
            database = table_attribute.split(".")[0]
            schema = table_attribute.split(".")[1]
            table = table_attribute.split(".")[-1]
        else:
            database = organization.get("database")
            schema = organization.get("schema")
            table = table_attribute
        metadata = {
            "database": database,
            "schema": schema,
            "table": table,
        }
        return metadata


from datetime import datetime
import hashlib
from more_itertools import lstrip
import pandas as pd
from typing import List, Optional
import uuid

def hash_string(input_string: str) -> int:
    """
    This function accepts a string and returns an 8 digit integer representation
    Method: generate a sha256 hash of the string and use modulo division to truncate it's integer representation to 8 digits

    WARNING:
    This function WILL return the same integer value for a string each time it is passed in the same session
    This function WILL NOT return a unique value for every unique string combination
    """
    return int(hashlib.sha256(input_string.encode('utf-8')).hexdigest(), 16) % 10**8

def hash_list(input_list: List[str]) -> int:
    """This function accepts a list of strings and returns an 8 digit integer representation"""
    input_string = ''.join(input_list)
    return hash_string(input_string)

def make_timestamp() -> str:
    now = datetime.now()
    return now.strftime("%Y_%m_%d_%H_%M")

def make_fqtn(table: str,
              database: Optional[str] = None,
              schema: Optional[str] = None
              ) -> str:
    """
    Accepts component parts and returns a possible fully qualified table string
    """
    if database is not None and schema is not None:
        return f"{database}.{schema}.{table}"
    if database is not None and schema is None:
        return f"{database}.PUBLIC.{table}"
    if database is None:
        return table

def parse_fqtn(fqtn: str, org_defaults: dict = None) -> tuple:
    """
    Accepts a possible fully qualified table string and returns its component parts
    """
    database = org_defaults.get('database') if org_defaults else None
    schema = org_defaults.get('schema') if org_defaults else None
    table = fqtn
    if fqtn.count('.') == 2:
        database = fqtn.split(".")[0]
        schema = fqtn.split(".")[1]
        table = fqtn.split(".")[-1]
    elif fqtn.count('.') == 1:
        schema = fqtn.split(".")[0]
        table = fqtn.split(".")[-1]
    return database, schema, table

def random_table_name() -> str:
    identifier = str(uuid.uuid4())
    table_name = ''.join(list(lstrip(hashlib.md5(identifier.encode('utf-8')).hexdigest(),
                            lambda x: x.isnumeric()))).upper()
    return table_name

def _cleanse_sql_data_type(dtype: str) -> str:
    """
    Converts a string to Snowflake compliant data type
    """
    if dtype.lower() in ["object", "text", "variant"]:
        return "string"
    else:
        return dtype.lower()

def _cleanse_sql_list(list_in: List[str]) -> List[str]:
    """
    Converts a list of strings to Snowflake compliant names
    """
    return [_cleanse_sql_name(n) for n in list_in]

def _cleanse_sql_name(name: str) -> str:
    """
    Converts a string to a snowflake compliant value
    """
    return name.replace(" ", "_").replace("-", "_").replace('"', '').replace(".","_").upper()

def _is_restricted_sql(sql: str) -> bool:
    """
    Checks a SQL string for presence of dangerous keywords
    """
    if any(word in sql.upper() for word in SNOWFLAKE_RESTRICTED_KEYWORDS):
        return True
    return False

def _is_scary_sql(sql: str) -> bool:
    """
    Checks a SQL string for presence of injection keywords
    """
    if any(word in sql.upper() for word in SQL_INJECTION_KEYWORDS):
        return True
    return False

def _is_valid_view_sql(sql: str) -> bool:
    """
    Checks a SQL string for presence of structural keywords
    """
    mandatory_words = ["SELECT", "FROM"]
    if not all(word in sql.upper() for word in mandatory_words):
        return False
    return True

def _quote_restricted_keywords(keyword: str) -> str:
    if keyword in SNOWFLAKE_RESTRICTED_KEYWORDS:
        return f'"{keyword}"'
    return keyword


SQL_RESTRICTED_CHARACTERS = [
    ' ', '-', ';'
]

SQL_INJECTION_KEYWORDS = [
    'DELETE',
    'TRUNCATE',
    'DROP',
    'ALTER',
    'UPDATE',
    'INSERT',
    'MERGE'
]

SNOWFLAKE_RESTRICTED_KEYWORDS = [
    'ACCOUNT','ALL','ALTER','AND','ANY','AS','BETWEEN','BY',
    'CASE','CAST','CHECK','COLUMN','CONNECT','CONNECTION','CONSTRAINT',
    'CREATE','CROSS','CURRENT','CURRENT_DATE','CURRENT_TIME','CURRENT_TIMESTAMP',
    'CURRENT_USER','DATABASE','DELETE','DISTINCT','DROP','ELSE','EXISTS','FALSE',
    'FOLLOWING','FOR','FROM','FULL','GRANT','GROUP','GSCLUSTER','HAVING','ILIKE',
    'IN','INCREMENT','INNER','INSERT','INTERSECT','INTO','IS','ISSUE','JOIN','LATERAL',
    'LEFT','LIKE','LOCALTIME','LOCALTIMESTAMP','MINUS','NATURAL','NOT','NULL','OF','ON',
    'OR','ORDER','ORGANIZATION','QUALIFY','REGEXP','REVOKE','RIGHT','RLIKE','ROW','ROWS',
    'SAMPLE','SCHEMA','SELECT','SET','SOME','START','TABLE','TABLESAMPLE','THEN','TO','TRIGGER',
    'TRUE','TRY_CAST','UNION','UNIQUE','UPDATE','USING','VALUES','VIEW','WHEN','WHENEVER','WHERE','WITH'
]


from typing import List, Optional

from .error import APIError
from pyrasgo.utils import track_usage

class Delete():

    def __init__(self):
        from .connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    @track_usage
    def collection(self, id: int):
        """
        Permanently delete a Rasgo Collection. There is no way to undo this operation.
        """
        try:
            response = self.api._delete(f"/models/{id}", api_version=1)
            if response.status_code == 200:
                return f"Collection {id} successfully deleted"
            if response.status_code == 403:
                raise APIError(f"User does not have access to delete Collection {id}")
            raise APIError(f"Problem deleting Collection {id}.")
        except:
            raise APIError(f"Problem deleting Collection {id}.")

    @track_usage
    def data_source(self, id: int):
        """
        Permanently delete a Rasgo DataSource. There is no way to undo this operation.
        """
        try:
            response = self.api._delete(f"/data-source/{id}", api_version=1)
            if response.status_code == 200:
                return f"DataSource {id} successfully deleted"
            if response.status_code == 403:
                raise APIError(f"User does not have access to delete DataSource {id}")
            return f"Problem deleting DataSource {id}."
        except:
            return f"Problem deleting DataSource {id}."

    @track_usage
    def feature(self, id: int):
        """
        Permanently delete a Rasgo Feature. There is no way to undo this operation.
        """
        try:
            response = self.api._delete(f"/features/{id}", api_version=1)
            if response.status_code == 200:
                return f"Feature {id} successfully deleted"
            if response.status_code == 403:
                raise APIError(f"User does not have access to delete Feature {id}")
            return f"Problem deleting Feature {id}."
        except:
            return f"Problem deleting Feature {id}."


from typing import Dict, List, Optional, Union

import pandas as pd
import time
from tqdm import tqdm

from pyrasgo.api.connection import Connection
from pyrasgo.schemas import data_source as schema
from pyrasgo.utils import naming, track_usage

class DataSource(Connection):
    """
    Stores a Rasgo DataSource
    """

    def __init__(self, api_object, **kwargs):
        super().__init__(**kwargs)
        self._fields = schema.DataSource(**api_object)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Source(id={self.id}, name={self.name}, sourceType={self.sourceType}, table={naming.make_fqtn(self.table, self.tableDatabase, self.tableSchema)})"

    def __getattr__(self, item):
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            self.refresh()
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

# ----------
# Properties
# ----------
    @property
    @track_usage
    def columns(self) -> List[schema.DataSourceColumn]:
        """
        Returns columns in the DataSource table
        """
        response = self._get(f"/data-source/columns/{self.id}", api_version=1).json()
        return [schema.DataSourceColumn(**column) for column in response['columns']]

# -------
# Methods
#--------
    @track_usage
    def display_source_code(self):
        """
        Convenience function to display the sourceCode property
        """
        return self._fields.sourceCode

    @track_usage
    def read_into_df(self,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        """
        Pull Source data from DataWarehouse into a pandas Dataframe
        """
        from pyrasgo.api.read import Read
        return Read().source_data(id=self.id, filters=filters, limit=limit)

    @track_usage
    def rebuild_from_source_code(self):
        """
        Rebuild the Source using the source code
        """
        raise NotImplementedError()

    @track_usage
    def refresh(self):
        """
        Updates the Soure's attributes from the API
        """
        self._fields = schema.DataSource(**self._get(f"/data-source/{self.id}", api_version=1).json())

    @track_usage
    def rename(self, new_name: str):
        """
        Updates a DataSource's display name
        """
        print(f"Renaming DataSource {self.id} from {self.name} to {new_name}")
        source = schema.DataSourceUpdate(id=self.id, name=new_name)
        self._fields = schema.DataSource(**self._patch(f"/data-source/{self.id}",
                                                    api_version=1, _json=source.dict(exclude_unset=True, exclude_none=True)).json())

    def _make_table_metadata(self):
        organization = self._get_profile().get("organization")
        metadata = {
            "database": self._fields.tableDatabase or organization.get("database"),
            "schema": self._fields.tableSchema or organization.get("schema"),
            "table": self._fields.table,
        }
        return metadata


import datetime
import numpy as np
import pandas as pd
import sys
from tqdm import tqdm
from typing import List, Optional
import webbrowser

from pyrasgo.api.error import APIError
from pyrasgo.schemas.feature import featureImportanceStats, ColumnProfiles
from pyrasgo.storage.dataframe.utils import set_df_id, map_pandas_df_type, DataframeDataTypes
from pyrasgo.utils import track_experiment, track_usage

class Evaluate():

    def __init__(self, experiment_id = None):
        from pyrasgo.api import Save
        from pyrasgo.api.session import Environment

        self._experiment_id = experiment_id
        self._environment = Environment.from_environment()
        self.save = Save()

    @track_usage
    def duplicate_rows(self, df: pd.DataFrame, 
                       columns: List[str] = None) -> pd.DataFrame:
        """ 
        Returns a DataFrame of rows that are duplicated in your original DataFrame 
        
        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            columns: (Optional) List[str]:
                List of column names to check for duplicates in
        
        Returns
        -------
            pandas DataFrame
        """
        df_out = df.copy(deep=True)
        if columns:
            df_out = df.iloc[:0].copy()
            for column in columns:
                df_out = df_out.append(df[df.duplicated([column])])
        else:
            df_out = df[df.duplicated()]
        return df_out

    @track_usage
    @track_experiment
    def feature_importance(self, df: pd.DataFrame, 
                           target_column: str,
                           timeseries_index: Optional[str] = None,
                           exclude_columns: List[str] = None,
                           return_cli_only: bool = False,
                           catboost_parameters: dict = {},
                           _model_id: Optional[int] = None
                           ) -> dict:
        """
        Calculates importance of a target feature using Shapley values. 
        Uses an 80% training sample of data.
        Opens interactive graph in the Rasgo WebApp or returns raw json of feature importance.

        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            target_column: str: 
                Column name of target feature
            timeseries_index: (Optional) str:
                Name of column to use as timeseries index. 
                If passed, rows will be ordered by column and split.
                If None is passed, rows will be split randomly.
            exclude_columns: (Optional) List[str]: 
                Column names of features to be filered out from profile
            return_cli_only: (Optional) bool:
                Instructs function to not open Rasgo WebApp and return json in the CLI only
            catboost_parameters: (Optional) dict:
                Allows for control over the CatBoost modeling process. Arguments passed in this dictionary
                are used in a call to a CatBoost Regressor or Classifier. Only select hyperparameters 
                are supported.
            _model_id: (Optional) int:
                Collection ID for which to calculate feature importance.
                If passed, feature stats will be stored.
                If nothing is passed, dataframe stats will be stored.
        
        Returns
        -------
            dict
        """
        # Check if we can run this
        catboost_args = catboost_parameters.copy()
        try:
            import shap
            import catboost
        except ModuleNotFoundError:
            raise APIError('Missing dependencies needed to run this function. '
                           'Run `pip install pyrasgo[df]` to install shap and catboost packages.')

        if target_column not in df.columns:
            raise APIError(f'Column {target_column} does not exist in DataFrame')
        if timeseries_index and timeseries_index not in df.columns:
            raise APIError(f'Column {timeseries_index} does not exist in DataFrame')

        # Start progress bar
        with tqdm(desc='Calculating Feature Importance', total=8, unit='step', file=sys.stdout, leave=False) as progress:
            
            # assign a unique id to the DF before we alter it
            rasgo_df_id = set_df_id(df, self._experiment_id)
            progress.update()

            # Run profile for df
            timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            self.profile(df, exclude_columns=exclude_columns, return_cli_only=True, timestamp_override=timestamp)
            progress.update()

            # Copy df so we can alter it without impacting source data
            fi_df = df.copy(deep=True)
            progress.update()
            print(f"Total records in dataset before train-test split and null handling: {len(fi_df.index)}")

            # check distinct values of target variable
            # if 2, train class
            # if >2, run regressor

            target_distinct_vals = len(df[target_column].dropna().unique())
            if target_distinct_vals < 2:
                raise ValueError(f"The chosen target column ({target_column}) " \
                            f"has less than two distinct values. Please choose a " \
                            f"target that can have a regressor or classifier " \
                            f"trained against it.")
            progress.update()

            # Split data into train, test sets
            train_df, test_df = self.train_test_split(fi_df, training_percentage=0.8, timeseries_index=timeseries_index)            
            if timeseries_index:
                train_df = train_df.drop(timeseries_index, 1)
                test_df = test_df.drop(timeseries_index, 1)
            progress.update()

            # Prep DataFrame
            if exclude_columns:
                if timeseries_index and timeseries_index in exclude_columns:
                    exclude_columns.remove(timeseries_index)
                if target_column in exclude_columns:
                    exclude_columns.remove(target_column)
                for col in exclude_columns:
                    if col not in train_df.columns:
                        raise APIError(f'Column {col} does not exist in DataFrame')
                    train_df = train_df.drop(col, 1)
                    test_df = test_df.drop(col, 1)
            
            # NOTE: Nulls cause a problem with the importance calc:
            print(f"Total records in training dataset before null handling: {len(train_df.index)}")
            print(f"Total records in test dataset before null handling: {len(test_df.index)}")
            train_df.dropna(axis=1, how='all', inplace=True)
            train_df.dropna(inplace=True)
            test_df.dropna(axis=1, how='all', inplace=True)
            test_df.dropna(inplace=True)
            # NOTE: Dates cause a problem with the importance cals:
            train_df = train_df.select_dtypes(exclude=['datetime'])
            test_df = test_df.select_dtypes(exclude=['datetime'])
            print(f"Total records in training dataset after null handling: {len(train_df.index)}")
            print(f"Total records in test dataset after null handling: {len(test_df.index)}")

            # if there are no records in the train or test dataset, raise an
            # error and let the user know they have null issues
            if not len(train_df.index) and len(test_df.index):
                raise ValueError(f"After the train-test split and null " \
                    f"handling, the lengths of your train and test datasets " \
                    f"are {len(train_df.index)} and {len(test_df.index)}, " \
                    f"respectively. Please impute or otherwise handle your " \
                    f"null values if you'd like to run feature importance calculcations.")

            # Start building output json
            c_data = {}
            c_data["targetFeature"] = target_column

            # Create x and y df's based off target column
            train_x = train_df.loc[:, train_df.columns != target_column]
            train_y = train_df.loc[:, train_df.columns == target_column]
            test_x = test_df.loc[:, test_df.columns != target_column]
            test_y = test_df.loc[:, test_df.columns == target_column]

            # Get categorical feature indices to pass to catboost
            train_cat_features = np.where(train_x.dtypes != np.number)[0]
            test_cat_features = np.where(test_x.dtypes != np.number)[0]
            progress.update()

            # Create the catboost dataset
            try:
                train_dataset = catboost.Pool(data=train_x, label=train_y, cat_features=train_cat_features)
                test_dataset = catboost.Pool(data=test_x, label=test_y, cat_features=test_cat_features)
            except TypeError as e:
                raise APIError(f"Catboost error: {e}: "
                            f"One or more of the fields in your dataframe is a date that cannot be automatically filtered out. "
                            f"You can use the exclude_columns=[''] parameter to exclude these manually and re-run this fuction.")
            
            approved_catboost_params = [
            "iterations", "od_type", "od_pval", "od_wait", "use_best_model",
            "learning_rate", "max_depth", "loss_function", "train_dir"
            ]
            
            if catboost_args:
                # check to see if catboost parameter arguments all match the approved keys
                unsupported_params = [p for p in list(catboost_args.keys()) if p not in approved_catboost_params]
                if unsupported_params:
                    print(f"{approved_catboost_params} are the only " \
                        f"hyperparameters for pyrasgo users to control. Please limit customization to these parameters. " \
                        f"You passed in {unsupported_params}, which are unsupported."
                        )
                
                    # remove unsupported args from the catboost args 
                    catboost_args = {k:v for k,v in catboost_args.items() if k in approved_catboost_params}
            
            if not catboost_args.get("iterations"):
                # This defaults iterations to 300. If a user passes in iterations, we'll use that
                # if not, 300 is our baseline
                catboost_args["iterations"] = 300
            catboost_args["random_seed"] = 123

            # Build model and measure performance
            if target_distinct_vals > 2:
                catboost_args["custom_metric"] = ['R2','RMSE','MAPE']
                model = catboost.CatBoostRegressor(**catboost_args)
                try:
                    target_data_type = DataframeDataTypes[str(train_df[target_column].dtypes).upper()].value
                except(AttributeError, KeyError):
                    target_data_type = "string"
                if target_data_type != 'number':
                    print(f"Your target column has {target_distinct_vals} distinct values, but is not a numeric type. " \
                    f"Because there are more than two distinct values, we are running a regressor model on your data to " \
                    f"determine importance. If this behavior is unintened, please either change the type of your data or " \
                    f"change your target column.")
            elif target_distinct_vals == 2:
                catboost_args["custom_metric"] = ['AUC','Logloss','Precision','Recall']
                model = catboost.CatBoostClassifier(**catboost_args)

            model.fit(train_dataset, eval_set=test_dataset, use_best_model=True, verbose=False, plot=False)
            metrics = model.get_best_score()
            learn_set = metrics.get('learn')
            validation_set = metrics.get('validation')

            # NOTE: learn stats will not be included in response model, but are here for future use
            c_data['modelPerformance'] = {}
            c_data['modelPerformance']['learn'] = {}
            c_data['modelPerformance']['learn']['RMSE'] = learn_set.get('RMSE')
            c_data['modelPerformance']['learn']['R2'] = learn_set.get('R2')
            c_data['modelPerformance']['learn']['MAPE'] = learn_set.get('MAPE')
            c_data['modelPerformance']['learn']['AUC'] = learn_set.get('AUC')
            c_data['modelPerformance']['learn']['LogLoss'] = learn_set.get('LogLoss')
            c_data['modelPerformance']['learn']['Precision'] = learn_set.get('Precision')
            c_data['modelPerformance']['learn']['Recall'] = learn_set.get('Recall')
            c_data['modelPerformance']['validation'] = {}
            c_data['modelPerformance']['validation']['RMSE'] = validation_set.get('RMSE')
            c_data['modelPerformance']['validation']['R2'] = validation_set.get('R2')
            c_data['modelPerformance']['validation']['MAPE'] = validation_set.get('MAPE')
            c_data['modelPerformance']['validation']['AUC'] = validation_set.get('AUC')
            c_data['modelPerformance']['validation']['Logloss'] = validation_set.get('Logloss')
            c_data['modelPerformance']['validation']['Precision'] = validation_set.get('Precision')
            c_data['modelPerformance']['validation']['Recall'] = validation_set.get('Recall')
            progress.update()

            # Calculate shapley values for feature importance
            explainer = shap.TreeExplainer(model)
            shap_values = explainer.shap_values(train_dataset)
            df_shap = pd.DataFrame(shap_values, columns=train_x.columns)

            # Calculate alternative feature importance for comparison
            df_comp = pd.DataFrame({'feature_importance': model.get_feature_importance(train_dataset), 
                                    'feature_names': train_x.columns
                                    }).sort_values(by=['feature_importance'], ascending=False)            
            progress.update()

            # Histogram binning of shapley values
            c_data['featureShapleyDistributions'] = {}
            for column in df_shap:
                try:
                    b = min(10, train_df[column].nunique())
                    H, xedges, yedges = np.histogram2d(x=train_df[column], y=df_shap[column], bins=b)
                    df_hist = pd.DataFrame(zip(H.tolist(), xedges, yedges), columns=['Histogram','feature_edges','shap_edges'])
                    fhist = df_hist.to_dict(orient="list")
                    c_data['featureShapleyDistributions'][column] = fhist
                except:
                    list_of_counts = []
                    list_of_values = []
                    col_min = np.nanmin(df_shap[column])
                    col_max = np.nanmax(df_shap[column])
                    unique_values = train_df[column].value_counts().index.astype(str).tolist()
                    if len(unique_values) > 10:
                        top_10_values = unique_values[:10]
                    else:
                        top_10_values = unique_values
                    df_combined = pd.concat([train_df[column], df_shap[column]], axis=1, join='inner')
                    shapcol = column+"_shap"
                    df_combined.columns = [column, shapcol]
                    for val in top_10_values:
                        count, division = np.histogram(df_combined.query('{0}==@val'.format(column))[shapcol], bins=10, range=(col_min, col_max), density=False)
                        list_of_counts.append(count.tolist())
                        list_of_values.append(val)
                    df_hist = pd.DataFrame(zip(list_of_counts, list_of_values, division), columns=['Histogram','feature_edges','shap_edges'])
                    fhist = df_hist.to_dict(orient="list")
                    c_data['featureShapleyDistributions'][column] = fhist
            progress.update()
            
            # Mean absolute value by feature
            # NOTE: default importance stats will not be included in response model, but are here for future use
            c_data['featureImportance'] = {}
            c_data['featureImportance']['shapley'] = df_shap.abs().mean().to_dict()
            #c_data['featureImportance']['default'] = df_comp.abs().mean().to_dict()

            # Prepare the response
            response = {
                "targetFeature": target_column,
                "featureShapleyDistributions": c_data['featureShapleyDistributions'],
                "featureImportance": c_data['featureImportance']['shapley'],
                "modelPerformance": c_data['modelPerformance']['validation'],
                "timestamp": timestamp
            }

            if _model_id:
                response["modelId"] = _model_id
                url = f"{self._environment.app_path}/collections/{_model_id}"
                response['url'] = url
                json_payload = featureImportanceStats(**response)
                self.save.feature_importance_stats(id=_model_id, payload=json_payload)
            else:
                url = f'{self._environment.app_path}/dataframes/{rasgo_df_id}/importance'
                response['url'] = url
                json_payload = featureImportanceStats(**response)
                self.save.column_importance_stats(id=rasgo_df_id, payload=json_payload)

            print('Importance URL:', url)
            if not return_cli_only:
                webbrowser.open(url)
            
            progress.close()
        return response

    @track_usage
    def missing_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """ 
        Print all columns in a Dataframe with null values
        and return rows with null values
        
        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
        
        Returns
        -------   
            'Columns with null values:'
            '-------------------------'
            'column, count of rows'
            '-------------------------'
            'List[index of rows with null values]'
        """
        column_with_nan = df.columns[df.isnull().any()]
        template="%-20s %-6s"
        print(template % ("Column", "Count of Nulls"))
        print("-"*35)
        for column in column_with_nan:
            print(template % (column, df[column].isnull().sum()))
        print("-"*35)
        return df[df.isnull().any(axis=1)]

    @track_usage
    def profile(self, df: pd.DataFrame, 
                exclude_columns: List[str] = None,
                return_cli_only: bool = False,
                timestamp_override: str = None) -> dict:
        """
        Profile a DataFrame locally, and push metadata to rasgo to display.

        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            exclude_columns: List[str]: 
                Column names of features to be filered out from profile
            return_cli_only: (Optional) bool:
                Instructs function to not open Rasgo WebApp and return json in the CLI only
        
        Returns
        -------
            dict
        """
        # NOTE: List of optimizations post-mvp:
        # - add frequency and outlier data to commonValues

        # assign a unique id to the DF before we alter it
        rasgo_df_id = set_df_id(df, self._experiment_id)

        # Get timestamp
        timestamp = timestamp_override or datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # Copy df so we can alter it without impacting source data
        p_df = df.copy(deep=True)
        # Remove columns before profiling
        if exclude_columns:
            for col in exclude_columns:
                if col not in p_df.columns:
                    raise APIError(f'Column {col} does not exist in DataFrame')
                p_df = p_df.drop(col, 1)

        # Create an object to hold intermediate data calculated via pandas
        profile_data = {}
        for col_label in p_df:
            profile_data[col_label] = {}

        # Calculate count stats before dropping nulls
        for col_label in p_df:
            profile_data[col_label]["recCt"] = p_df[col_label].count() + p_df[col_label].isna().sum()
            profile_data[col_label]["distinctCt"] = p_df[col_label].nunique()
            profile_data[col_label]["nullRecCt"] = p_df[col_label].isna().sum()
            profile_data[col_label]["zeroValRecCt"] = (p_df[col_label] == 0).astype(int).sum()


        print(f"Profile - total records in dataset before null handling: {len(p_df.index)}")
        p_df.dropna(axis=1, how='all', inplace=True)
        p_df.dropna(inplace=True)
        print(f"Profile - total records in dataset after null handling: {len(p_df.index)}")

        if not len(p_df.index):
            raise ValueError(f"Null record handling in dataframe profiling removed all " \
                f"the records from your dataframe. Please impute or otherwise" \
                f" handle your null values in order to run dataframe profiling.")

        # Boolean columns can't calculate quantiles, but if we represent them
        # as 1 and 0, they can. We then store the name of the columns to
        # return them to boolean types once the stats have been calculated.
        casted_bools = []
        for col_label in p_df:
            if p_df[col_label].dtype.name == "bool":
                p_df[col_label] = p_df[col_label].astype(int)
                casted_bools.append(col_label)

        # label is the name of the attribute in the api request/response
        # mean is the name of the pandas function that calculates it
        # last item is any kwargs required for the function
        labels_and_functions = [
            ('meanVal', 'mean', {'axis': 0, 'numeric_only': True}),
            ('medianVal', 'median', {'axis': 0, 'numeric_only': True}),
            ('maxVal', 'max', {'axis': 0, 'numeric_only': True}),
            ('minVal', 'min', {'axis': 0, 'numeric_only': True}),
            ('sumVal', 'sum', {'axis': 0, 'numeric_only': True}),
            ('stdDevVal', 'std', {'axis': 0, 'numeric_only': True}),
            ('varianceVal', 'var', {'axis': 0, 'numeric_only': True}),
            ('skewVal', 'skew', None),
            ('kurtosisVal', 'kurtosis', None),
            ('q1Val', 'quantile', {'q': .25, 'axis': 0, 'numeric_only': True}),
            ('q3Val', 'quantile', {'q': .75, 'axis': 0, 'numeric_only': True}),
            ('pct5Val', 'quantile', {'q': .05, 'axis': 0, 'numeric_only': True}),
            ('pct95Val', 'quantile', {'q': .95, 'axis': 0, 'numeric_only': True})
        ]
        for label, func_name, extra_args in labels_and_functions:
            results = self._evaluate_df(p_df, func_name, extra_args)
            for col_label in p_df:
                try:
                    # store under key name for column, if the value exists
                    profile_data[col_label][label] = results[col_label]
                except KeyError:
                    pass
        
        # change boolean columns back to bool
        for col_label in casted_bools:
            p_df[col_label] = df[col_label].astype(bool)

        # Warning: any data types not in the DataframeDataTypes enum will
        # default to string in the below logic. New data types will need
        # to be added to the enum
        # Generate values histogram
        histo_data = {}
        common_values = {}
        for col_label in p_df:
            # Set column data type and compare against enum
            try:
                col_type = DataframeDataTypes[str(p_df[col_label].dtypes).upper()].value
            except (AttributeError, KeyError):
                col_type = "string"
            # Histogram for numeric
            if col_type == "number":
                counts, edges = np.histogram(p_df[col_label], bins='auto', density=False)
                df_hist = pd.DataFrame(zip(counts, edges), columns=['height','bucketFloor'])
                df_hist['bucketCeiling']  = df_hist['bucketFloor']
                fhist = df_hist.to_dict(orient="records")
                # Common values for numeric
                ch = p_df[col_label].value_counts().rename_axis('val').reset_index(name='recCt')
                top5 = ch.head(10).to_dict(orient="records")
                total_rec = profile_data[col_label]["recCt"]
                for entry in top5:
                    entry["freq"] = entry["recCt"]/total_rec

            elif col_type == "string":
                p_df[col_label] = p_df[col_label].astype(str)
                dh = p_df[col_label].apply(str).value_counts().rename_axis('bucketFloor').reset_index(name='height')
                dh['bucketCeiling'] = dh['bucketFloor']
                fhist = []
                ch = dh.rename(columns={"bucketFloor":"val", "height":"recCt"}).drop(columns=["bucketCeiling"])
                top5 = ch.head(10).to_dict(orient="records")
                total_rec = profile_data[col_label]["recCt"]
                for entry in top5:
                    entry["freq"] = entry["recCt"]/total_rec

            # TODO: currently handling any non-numeric or string types as strings. Expand later
            else:
                p_df[col_label] = p_df[col_label].astype(str)
                dh = p_df[col_label].apply(str).value_counts().rename_axis('bucketFloor').reset_index(name='height')
                dh['bucketCeiling'] = dh['bucketFloor']
                fhist = []
                ch = dh.rename(columns={"bucketFloor":"val", "height":"recCt"}).drop(columns=["bucketCeiling"])
                top5 = ch.head(10).to_dict(orient="records")
                total_rec = profile_data[col_label]["recCt"]
                for entry in top5:
                    entry["freq"] = entry["recCt"]/total_rec

            histo_data[col_label] = fhist
            common_values[col_label] = top5

        # Prepare response
        response = {
            'columnProfiles': [],
            'timestamp': timestamp
        }
        for col_label in profile_data:
            column_stats = {
                'columnName': col_label,
                'dataType': map_pandas_df_type(p_df[col_label].dtype.name),
                'featureStats': profile_data[col_label],
                'commonValues': common_values[col_label]
            }
            if histo_data[col_label]:
                column_stats["histogram"] = histo_data[col_label]
            response['columnProfiles'].append(column_stats)

        url = f'{self._environment.app_path}/dataframes/{rasgo_df_id}/features'
        response['url'] = url
        json_payload = ColumnProfiles(**response)
        self.save.dataframe_profile(id=rasgo_df_id, payload=json_payload)

        if not return_cli_only:
            # open the feature profiles in a web page
            webbrowser.open(url)
            print('Profile URL:', url)
        return response

    @track_usage
    def timeseries_gaps(self, df: pd.DataFrame, 
                        datetime_column: str,
                        partition_columns: List[str] = []) -> pd.DataFrame:
        """
        Returns a dataframe of rows before and after timeseries gaps
        
        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            datetime_column: str: 
                Name of column to check for timeseries gaps
            partition_columns: List[str]:
                Column names to group by

        Returns
        -------
            pandas DataFrame
        """
        # NOTE: List of optimizations post-mvp:
        # - Check that datetime_column can convert to_datetime without error
        # - Support non-date grains

        # Check for columns
        sort_columns = partition_columns.copy()
        sort_columns.append(datetime_column)
        for col in sort_columns:
            if col not in df.columns:
                raise APIError(f'Column {col} does not exist in DataFrame')
        group_by_columns = datetime_column if partition_columns == [] else partition_columns

        # Calculate lags and sort
        ts_df = df.copy(deep=True)
        ts_df.sort_values(by=sort_columns, inplace=True)
        ts_df['TSGAPDateCol'] = pd.to_datetime(ts_df[datetime_column], infer_datetime_format=True)
        ts_df['TSGAPLastDate'] = ts_df.groupby(group_by_columns)['TSGAPDateCol'].shift(1)
        ts_df['TSGAPNextDate'] = ts_df.groupby(group_by_columns)['TSGAPDateCol'].shift(-1)

        # NOTE: This assumes a day grain in the timeseries column, we'll need to expand this to support more grains 
        # Select the rows for output
        df_out = ts_df[(ts_df['TSGAPNextDate'] - ts_df['TSGAPDateCol'] != '1 days') 
                     | (ts_df['TSGAPDateCol'] - ts_df['TSGAPLastDate'] != '1 days')]
        df_out.drop(['TSGAPDateCol'], axis=1, inplace=True)
        #df_out.drop(['TSGAPNextDate', 'TSGAPLastDate'], axis=1, inplace=True)

        return df_out

    @track_usage
    @track_experiment
    def train_test_split(self, df: pd.DataFrame, 
                         training_percentage: float = .8,
                         timeseries_index: Optional[str] = None) -> pd.DataFrame:
        """
        Splits a single dataframe into training and test dataframes based on input percentage
        
        Timeseries: orders rows by input datetime index and returns a precise split
        Non-Timeseries: randomly orders rows and returns an approximate split

        Parameters
        ----------
            df: pandas DataFrame:
                    Dataframe to operate on
            training_percentage: float: 
                Percentage of rows to use as training set (default = 80%)
            timeseries_index: (Optional) str: 
                Name of column to use as timeseries index. 
                If passed, rows will be ordered by column and split.
                If None is passed, rows will be split randomly.
        
        Returns
        -------
            pandas DataFrame, pandas DataFrame
        """
        if timeseries_index is None:
            # Non-timeseries data: Split randomly
            rng = np.random.default_rng(1991)
            msk = rng.random(len(df)) <= training_percentage
            train_df = df[msk]
            test_df = df[~msk]
        
        else:
            # Timeseries data: Split in date order
            if timeseries_index not in df.columns:
                raise APIError(f'Column {timeseries_index} does not exist in DataFrame')

            # Validation check for Datetime column
            if df[timeseries_index].dtype not in [np.datetime64, 'datetime64[ns]']:
                print(f'The column you passed as timeseries_index is {df[timeseries_index].dtype}, which is not a valid datetime type. '
                      f'This may results in incorrectly building the test set for your model.')

            # order the frame by date
            df.sort_values(by=[timeseries_index], inplace=True)

            # split into 2 frames based on training percentage
            row_ct = df.shape[0]
            train_ct = round(row_ct * training_percentage)
            test_ct = round(row_ct * (1-training_percentage))
            train_df = df.head(train_ct)
            test_df = df.tail(test_ct)
        return train_df, test_df

    @track_usage
    def type_mismatches(self, df: pd.DataFrame, 
                        column: str, 
                        data_type: str) -> pd.DataFrame:
        """ 
        Return a copy of your DataFrame with a column cast to another datatype
        
        Parameters
        ----------
            df: pandas DataFrame:
                Dataframe to operate on
            column: str:
                The column name in the DataFrame to cast
            data_type: str:
                The data type to cast to Accepted Values: ['datetime', 'numeric']
        
        Returns
        -------
            pandas DataFrame
        """
        new_df = pd.DataFrame()
        if data_type == 'datetime':
            new_df[column] = pd.to_datetime(df[column], errors='coerce', infer_datetime_format=True)
        elif data_type == 'numeric':
            new_df[column] = pd.to_numeric(df[column], errors='coerce')
        else:
            return "Supported data_type values are: 'datetime' or 'numeric'"
        total = df[column].count()
        cant_convert = new_df[column].isnull().sum()
        print(f"{(total - cant_convert) / total}%: {cant_convert} rows of {total} rows cannot convert.")
        new_df.rename(columns = {column:f'{column}CastTo{data_type.title()}'}, inplace = True)
        return new_df

    @classmethod
    def _evaluate_df(cls, df: pd.DataFrame, func_name: str, extra_args: dict):
        extra_args = extra_args or {}
        # Get the correct DataFrame function by name
        results = getattr(df, func_name)(**extra_args)
        return results



import pandas as pd
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
import webbrowser

from pyrasgo import config

from pyrasgo.api import Admin, Create, Delete, Get, Match, Publish, Read, Save, Update
from pyrasgo.primitives.feature import Feature, FeatureList
from pyrasgo.primitives.feature_set import FeatureSet
from pyrasgo.primitives.source import DataSource
from pyrasgo.primitives.collection import Collection

from pyrasgo import schemas as api
from pyrasgo.schemas.enums import Granularity, ModelType

from pyrasgo.storage  import Evaluate, Prune, Transform
from pyrasgo.storage.dataframe.utils import generate_unique_id

from pyrasgo.utils import ingestion
from pyrasgo.utils import track_usage


class Rasgo():
    """
    Base connection object to handle interactions with the Rasgo API.
    """
    def __init__(self):
        self._experiment_id = None
        self.get = Get()
        self.match = Match()
        self.create = Create()
        self.update = Update()
        self.publish = Publish()
        self.read = Read()
        self.delete = Delete()
        self.save = Save()
        self.transform = Transform()
        self.evaluate = Evaluate()
        self.prune = Prune()
        self.admin = Admin()

    def open_docs(self):
        webbrowser.open("https://docs.rasgoml.com/rasgo-docs/pyrasgo/pyrasgo-getting-started")

    def pronounce_rasgo(self):
        webbrowser.open("https://www.spanishdict.com/pronunciation/rasgo?langFrom=es")

    def _pyrasgo_version(self):
        from pyrasgo.version import __version__
        return __version__

    def _pyrasgo_api_key(self):
        from pyrasgo.config import get_session_api_key
        return get_session_api_key()

    @track_usage
    def activate_experiment(
        self,
        experiment_name: str
    ):
        """
        Activate an experiment with a unique name and id. Experiments are used to
        track dataframe activity across Rasgo methods

        Note: Experiment names must be unique
              Using an existing name will re-activate an old experiment
        """
        df = self.match.dataframe(name=experiment_name)
        adjective = 'existing'
        if not df:
            uid = generate_unique_id()
            df = self.save.dataframe(name=experiment_name, unique_id=uid)
            adjective = 'new'
        self._experiment_id = df.uniqueId
        self.evaluate = Evaluate(experiment_id=self._experiment_id)
        self.prune = Prune(experiment_id=self._experiment_id)
        print(f"Activated {adjective} experiment with name {experiment_name} for dataframe: {df.uniqueId}")


    @track_usage
    def end_experiment(self):
        """
        If an experiment is active, deactivate it and stop tracking dataframe activity
        """
        self._experiment_id = None
        self.evaluate = Evaluate()
        self.prune = Prune()
        print(f"Experiment ended")


    @track_usage
    def is_experiment_active(self):
        """
        Check if an experiment is active
        """
        print(f"Experiment active for dataframe: {self._experiment_id}" if self._experiment_id else "No experiment active")

# ---------
# Get Calls
# ---------
    @track_usage
    def get_collection(self, collection_id: int) -> Collection:
        """
        Returns a Rasgo Collection (set of joined Features) matching the specified id
        """
        return self.get.collection(collection_id)

    @track_usage
    def get_collection_attributes(self, collection_id: int) -> api.CollectionAttributes:
        """
        Returns a dict of attributes for a collection
        """
        return self.get.collection_attributes(collection_id)

    @track_usage
    def get_collections(self, include_shared: bool=False) -> List[Collection]:
        """
        Returns all Rasgo Collections (set of joined Features) that I have author access to. Add an include_shared
        parameter to return all Rasgo Collections that I have any access to (author or shared access)
        :param include_shared: Boolean value indicating if the return should include all accessible collections
        """
        return self.get.collections(include_shared=include_shared)

    @track_usage
    def get_collections_by_attribute(self, key: str, value: str = None) -> List[Collection]:
        """
        Returns a list of Rasgo Collections that match an attribute
        """
        return self.get.collections_by_attribute(key, value)

    @track_usage
    def get_columns_by_featureset(self, feature_set_id: int) -> List[api.Column]:
        """
        Returns all Columns in the specified FeatureSet
        """
        self.get.columns_by_featureset(feature_set_id)

    @track_usage
    def get_data_sources(self) -> List[DataSource]:
        """
        Returns all DataSources available in your organization or Rasgo Community
        """
        return self.get.data_sources()

    @track_usage
    def get_data_source(self, data_source_id: int) -> DataSource:
        """
        Returns the DataSource with the specified id
        """
        return self.get.data_source(data_source_id)

    @track_usage
    def get_data_source_columns(self, data_source_id: int) -> List[api.DataSourceColumn]:
        """
        Returns columns in the DataSource with the specified id
        """
        return self.get.data_source_columns(data_source_id)

    @track_usage
    def get_data_source_stats(self, data_source_id: int):
        """
        Returns the stats profile of the specificed data source
        """
        return self.get.data_source_stats(data_source_id)

    @track_usage
    def get_dimensionalities(self) -> List[api.Dimensionality]:
        """
        Returns all Dimensionalities available in your organization or Rasgo Community
        """
        return self.get.dimensionalities()

    @track_usage
    def get_feature(self, feature_id: int) -> Feature:
        """
        Returns the Feature with the specified id
        """
        return self.get.feature(feature_id)

    @track_usage
    def get_feature_attributes(self, feature_id: int) -> api.FeatureAttributes:
        """
        Returns a dict of attributes for a feature
        """
        return self.get.feature_attributes(feature_id)

    @track_usage
    def get_feature_attributes_log(self, feature_id: int) -> tuple:
        """
        Returns a list of all attributes values logged to a feature over time
        """
        return self.get.feature_attributes_log(feature_id)

    @track_usage
    def get_feature_set(self, feature_set_id: int) -> FeatureSet:
        """
        Returns the FeatureSet (set of Fetures) with the specified id
        """
        return self.get.feature_set(feature_set_id)

    @track_usage
    def get_feature_sets(self) -> List[FeatureSet]:
        """
        Returns a list of FeatureSets (set of Features) available in your organization or Rasgo Community
        """
        return self.get.feature_sets()

    @track_usage
    def get_feature_stats(self, feature_id: int) -> Optional[api.FeatureStats]:
        """
        Returns the stats profile for the specified Feature
        """
        return self.get.feature_stats(feature_id)

    @track_usage
    def get_features(self) -> FeatureList:
        """
        Returns a list of Features available in your organization or Rasgo Community
        """
        return self.get.features()

    @track_usage
    def get_features_by_attribute(self, key: str, value: str = None) -> List[Feature]:
        """
        Returns a list of features that match an attribute
        """
        return self.get.features_by_attribute(key, value)

    @track_usage
    def get_features_by_featureset(self, feature_set_id) -> FeatureList:
        """
        Returns a list of Features in the specific FeatureSet
        """
        return self.get.features_by_featureset(feature_set_id)

    @track_usage
    def get_shared_collections(self) -> List[Collection]:
        """
        Returns all Rasgo Collections (set of joined Features) shared in my organization or in Rasgo community
        """
        return self.get.shared_collections()

    @track_usage
    def get_source_columns(self, table: Optional[str] = None, database: Optional[str] = None, schema: Optional[str] = None, data_type: Optional[str] = None) -> pd.DataFrame:
        """
        Returns a DataFrame of columns in Snowflake tables and views that are queryable as feature sources
        """
        return self.get.source_columns(table, database, schema, data_type)

    @track_usage
    def get_source_tables(self, database: Optional[str] = None, schema: Optional[str] = None) -> pd.DataFrame:
        """
        Return a DataFrame of Snowflake tables and views that are queryable as feature sources
        """
        return self.get.source_tables(database, schema)


# -------------------
# Post / Create Calls
# -------------------
    @track_usage
    def create_collection(self, name: str,
                          type: Union[str, ModelType],
                          granularity: Union[str, Granularity],
                          description: Optional[str] = None,
                          is_shared: Optional[bool] = False) -> Collection:
        return self.create.collection(name, type, granularity, description, is_shared)

    @track_usage
    def put_collection_attributes(self, collection_id: int, attributes: List[dict]):
        """
        Create or update attributes on a Rasgo Collection

        param attributes: dict [{"key": "value"}, {"key": "value"}]
        """
        return self.update.collection_attributes(collection_id, attributes)

    @track_usage
    def put_feature_attributes(self, feature_id: int, attributes: List[dict]):
        """
        Create or update attributes on a feature

        param attributes: dict [{"key": "value"}, {"key": "value"}]
        """
        return self.update.feature_attributes(feature_id, attributes)


# ------------------
# Workflow Functions
# ------------------
    @track_usage
    def prepare_feature_set_dict(self, feature_set_id: int = None):
        """
        Returns a dict of metadata values for the specificed Rasgo FeatureSet

        params
        ------
        feature_set_id: int: ID of a Rasgo FeatureSet

        Alternate Usage
        ---------------
        Pass in no params to return a blank featureset dict template
        """
        if feature_set_id:
            fs = self.get.feature_set(feature_set_id)
            return ingestion.save_feature_set_to_dict(fs)
        return ingestion.RASGO_DICT

    @track_usage
    def prepare_feature_set_yml(self, feature_set_id: int, file_name: str, directory: str = None):
        """
        Saves a yml file of metadata values for the specificed Rasgo FeatureSet to a directory location

        params
        ------
        feature_set_id: int: ID of a Rasgo FeatureSet
        file_name: str: name for the output file
        directory: str: full dir location for the output file (minus file_name)
        """
        fs = self.get.feature_set_yml(feature_set_id)
        return ingestion.save_feature_set_to_yaml(fs, file_name=file_name, directory=directory)

    @track_usage
    def publish_features(self, features_dict: dict) -> FeatureSet:
        return self.publish.features(features_dict)

    @track_usage
    def publish_features_from_df(self,
                                 df: pd.DataFrame,
                                 dimensions: List[str],
                                 features: List[str],
                                 granularity: List[str] = [],
                                 tags: List[str] = [],
                                 sandbox: bool = True):
        raise NotImplementedError("`publish_features_from_df()` is a deprecated method. "
                                  "Please use `publish.features()` or `publish.features_from_source()` instead")

    @track_usage
    def publish_features_from_source(self,
                                     data_source_id: int,
                                     features: List[str],
                                     dimensions: List[str],
                                     granularity: List[str] = [],
                                     tags: List[str] = [],
                                     feature_set_name: str = None,
                                     sandbox: bool = True,
                                     if_exists: str = 'fail') -> FeatureSet:
        """
        Publishes a FeatureSet from an existing DataSource table

        params:
            data_source_id: ID to a Rasgo DataSource
            features: List of column names that will be features
            dimensions: List of column names that will be dimensions
            granularity: List of grains that describe the form of the data. Common values are: 'day', 'week', 'month', 'second'
            tags: List of tags to be added to all features
            feature_set_name: Optional name for the FeatureSet (if not passed a random string will be assigned)

            if_exists:  fail - returns an error message if a featureset already exists against this table
                        return - returns the featureset without operating on it
                        edit - edits the existing featureset
                        new - creates a new featureset

        return:
            Rasgo FeatureSet
        """
        return self.publish.features_from_source(data_source_id, features, dimensions, granularity, tags, feature_set_name, sandbox, if_exists)

    @track_usage
    def publish_features_from_yml(self,
                                  yml_file: str,
                                  sandbox: Optional[bool] = True,
                                  git_repo: Optional[str] = None) -> FeatureSet:
        """
        Publishes metadata about a FeatureSet to Pyrasgo

        :param yml_file: Rasgo compliant yml file that describes the featureset(s) being created
        :param sandbox: Status of the features (True = 'Sandboxed' | False = 'Productionized')
        :param git_repo: Filepath string to these feature recipes in git
        :return: description of the featureset created
        """
        return self.publish.features_from_yml(yml_file, sandbox, git_repo)

    @track_usage
    def publish_source_data(self,
                            source_type: str,
                            file_path: Optional[Path] = None,
                            df: Optional[pd.DataFrame] = None,
                            table: Optional[str] = None,
                            table_database:Optional[str] = None,
                            table_schema: Optional[str] = None,
                            data_source_name: Optional[str] = None,
                            data_source_domain: Optional[str] = None,
                            data_source_table_name: Optional[str] = None, parent_data_source_id: Optional[int] = None,
                            if_exists: Optional[str] = 'fail'
                            ) -> DataSource:
        """
        Push a csv, Dataframe, or table to a Snowflake table and register it as a Rasgo DataSource (TM)

        NOTES: csv files will import all columns as strings

        params:
            source_type: Values: ['csv', 'dataframe', 'table']
            df: pandas DataFrame (only use when passing source_type = 'dataframe')
            file_path: full path to a file on your local machine (only use when passing source_type = 'csv')
            table: name of a valid Snowflake table in your Rasgo account (only use when passing source_type = 'table')
            table_database: Optional: name of the database of the table passed in 'table' param (only use when passing source_type = 'table')
            table_schema: Optional: name of the schema of the table passed in 'table' param (only use when passing source_type = 'table')

            data_source_name: Optional name for the DataSource (if not provided a random string will be used)
            data_source_table_name: Optional name for the DataSource table in Snowflake (if not provided a random string will be used)
            data_source_domain: Optional domain for the DataSource (default is NULL)
            parent_data_source_id: Optional ID of a valid Rasgo DataSource that is a parent to this DataSource (default is NULL)

            if_exists: Values: ['fail', 'append', 'replace'] directs the function what to do if a DataSource already exists with this table name  (defaults to fail)

        return:
            Rasgo DataSource
        """
        return self.publish.source_data(source_type, file_path, df, table, table_database, table_schema, data_source_name, data_source_domain, data_source_table_name, parent_data_source_id, if_exists)

    @track_usage
    def read_collection_data(self,
                             collection_id: int,
                             filters: Optional[Dict[str, str]] = None,
                             limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Collection

        :param collection_id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        return self.read.collection_data(collection_id, filters, limit)

    @track_usage
    def read_feature_data(self,
                          feature_id: int,
                          filters: Optional[Dict[str, str]] = None,
                          limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Feature data

        :param feature_id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        return self.read.feature_data(feature_id, filters, limit)

    @track_usage
    def read_source_data(self,
                         source_id: int,
                         filters: Optional[Dict[str, str]] = None,
                         limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo DataSource

        :param source_id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        return self.read.source_data(source_id, filters, limit)


from pydantic import BaseModel
from typing import Optional
from pyrasgo.schemas.organization import Organization


class Dimensionality(BaseModel):
    id: int
    name: Optional[str]
    organizationId: Optional[str]


class DimensionalityCreate(BaseModel):
    dimensionType: str
    granularity: str


from requests.exceptions import HTTPError
from typing import List, Optional, Union

from .error import APIError
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, FeatureSet, DataSource
from pyrasgo.schemas.enums import Granularity, ModelType
from pyrasgo.schemas.feature import featureImportanceStats, ColumnProfiles
from pyrasgo.utils import track_usage


class Create():

    def __init__(self):
        from .connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    @track_usage
    def collection(self, name: str,
                   type: Union[str, ModelType],
                   granularity: Union[str, Granularity],
                   description: Optional[str] = None,
                   is_shared: Optional[bool] = False) -> Collection:
        try:
            # If not enum, convert to enum first.
            model_type = type.name
        except AttributeError:
            model_type = ModelType(type)

        try:
            # If not enum, convert to enum first.
            granularity = granularity.name
        except AttributeError:
            granularity = Granularity(granularity)

        content = {"name": name,
                   "type": model_type.value,
                   "granularities": [{"name": granularity.value}],
                   "isShared": is_shared
                   }
        if description:
            content["description"] = description
        response = self.api._post("/models", _json=content, api_version=1)
        return Collection(api_object=response.json())

    @track_usage
    def column(self, name: str,
               data_type: str,
               feature_set_id: int,
               dimension_id: int) -> api.Column:
        column = api.ColumnCreate(name=name, dataType=data_type,
                                  featureSetId=feature_set_id,
                                  dimensionId=dimension_id)
        response = self.api._post("/columns/", column.dict(exclude_unset=True), api_version=1).json()
        return api.Column(**response)

    @track_usage
    def data_source(self, table: str,
                    name: str,
                    source_type: str,
                    database: Optional[str] = None,
                    schema: Optional[str] = None,
                    source_code: Optional[str] = None,
                    domain: Optional[str] = None,
                    parent_source_id: Optional[int] = None) -> DataSource:
        data_source = api.DataSourceCreate(name=name,
                                           table=table,
                                           tableDatabase=database,
                                           tableSchema=schema,
                                           sourceCode=source_code,
                                           domain=domain,
                                           sourceType=source_type,
                                           parentId=parent_source_id)
        response = self.api._post("/data-source", data_source.dict(exclude_unset=True), api_version=1).json()
        return DataSource(api_object=response)

    @track_usage
    def dataframe(self, unique_id: str,
                  name: str = None,
                  shared_status: str = 'organization',
                  column_hash: Optional[str] = None,
                  update_date: str = None) -> api.Dataframe:
        shared_status = 'organization' if shared_status not in ['public', 'private'] else shared_status
        dataframe = api.DataframeCreate(uniqueId=unique_id,
                                        name=name,
                                        sharedStatus=shared_status,
                                        columnHash=column_hash,
                                        updatedDate=update_date)
        try:
            response = self.api._post("/dataframes", dataframe.dict(exclude_unset=True), api_version=1).json()
        except HTTPError as e:
            error_message = f"Failed to create dataframe {unique_id}."
            if e.response.status_code == 409:
                error_message += f" This id is already in use in your organization. Dataframe IDs must be unique."
            raise APIError(error_message)
        return api.Dataframe(**response)

    @track_usage
    def data_source_stats(self, data_source_id: int):
        """
        Sends an api request to build stats for a specified data source.
        """
        return self.api._post(f"/data-source/profile/{data_source_id}", api_version=1).json()

    @track_usage
    def dimensionality(self, granularity: str,
                       dimension_type: str) -> api.Dimensionality:
        """
        Creates a dimensionality record in a user's organization with format: DimensionType - Granularity
        """
        dimensionality = api.DimensionalityCreate(dimensionType=dimension_type,
                                                  granularity=granularity)
        response = self.api._post("/dimensionalities", dimensionality.dict(exclude_unset=True), api_version=1).json()
        return api.Dimensionality(**response)

    @track_usage
    def feature(self, feature_set_id: int,
                display_name: str,
                column_name: str,
                description: str,
                column_id: int,
                status: str,
                git_repo: str,
                tags: Optional[List[str]] = None) -> Feature:
        feature = api.FeatureCreate(name=display_name,
                                    code=column_name,
                                    description=description,
                                    featureSetId=feature_set_id,
                                    columnId=column_id,
                                    orchestrationStatus=status,
                                    tags=tags or [],
                                    gitRepo=git_repo)
        try:
            response = self.api._post("/features/", feature.dict(exclude_unset=True), api_version=1).json()
        except HTTPError as e:
            error_message = f"Failed to create Feature {display_name}."
            if e.response.status_code == 409:
                error_message += f" This name is already in use in your organization. Feature names must be unique."
            raise APIError(error_message)
        return Feature(api_object=response)

    @track_usage
    def feature_set(self, data_source_id: int,
                    name: str,
                    table_name: str,
                    file_path: Optional[str] = None,
                    source_code: Optional[str] = None) -> FeatureSet:
        feature_set = api.FeatureSetCreate(name=name,
                                           snowflakeTable=table_name,
                                           dataSourceId=data_source_id,
                                           rawFilePath=file_path,
                                           sourceCode=source_code)
        response = self.api._post("/feature-sets/", feature_set.dict(), api_version=0).json()
        return FeatureSet(api_object=response)

    @track_usage
    def feature_set_stats(self, feature_set_id: int):
        """
        Sends an api request to build feature stats for a specified feature.
        """
        return self.api._post(f"/feature-sets/{feature_set_id}/stats", api_version=1).json()

    @track_usage
    def feature_stats(self, feature_id: int):
        """
        Sends an api request to build feature stats for a specified feature.
        """
        return self.api._post(f"/features/{feature_id}/stats", api_version=1).json()

    @track_usage
    def feature_importance_stats(self, id: int, payload: featureImportanceStats):
        """
        Sends an api requrest to build feature importance stats for the specified model
        """
        return self.api._post(f"/models/{id}/stats/feature-importance", payload.dict(), api_version=1).json()

    @track_usage
    def column_importance_stats(self, id: str, payload: featureImportanceStats):
        """
        Sends a json payload of importance from a dataFrame to the API so it can render in the WebApp
        """
        return self.api._post(f"/dataframes/{id}/feature-importance", payload.dict(), api_version=1).json()

    @track_usage
    def dataframe_profile(self, id: str, payload: ColumnProfiles):
        """
        Send a json payload of a dataframe profile so it can render in the WebApp
        """
        return self.api._post(f"/dataframes/{id}/profile", payload.dict(), api_version=1).json()


import pandas as pd
from pathlib import Path
from types import FunctionType
from typing import List, Optional
import webbrowser
import yaml

from .error import APIError, ParameterValueError
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, FeatureSet, DataSource
from pyrasgo.schemas.enums import Granularity, ModelType
from pyrasgo.storage.dataframe import utils as dfutils
from pyrasgo.utils import ingestion, naming, track_usage


class Publish():

    def __init__(self):
        from . import Create, Get, Match, Save, Update
        from .connection import Connection
        from pyrasgo.config import get_session_api_key
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()
        self.get = Get()
        self.match = Match()
        self.create = Create()
        self.update = Update()
        self.save = Save()


    @track_usage
    def experiment(self, df: pd.DataFrame,
                         dimensions: List[str],
                         granularity: List[str],
                         features: List[str] = [],
                         return_cli_only: bool = True,
                         verbose: bool = True) -> Collection:
        """
        Accepts a dataframe and publishes:
        - DataFrame as Rasgo DataSource
        - Dataframe columns as Rasgo Features
        - Rasgo Collection comprised of all Features

        params
        ------
        df: DataFrame: Pandas DataFrame to use as source data
        features: List[str]: List of columns to add as features.
                  If none is passed, all columns not in `dimensions` list will be registered as features
        dimensions: List[str]: List of columns to add as dimensions
        granularity: List[str]: List of granularities to apply to dimensions
        verbose: boolean: Print status statements to stdout while function executes

        Returns Rasgo Collection
        """
        if verbose:
            print('Publishing DataFrame as a Rasgo DataSource')
        try:
            source = self.source_data(source_type='dataframe', df=df, verbose=verbose)
            if verbose:
                print('Successfully created:', source)
        except Exception as e:
            print(f'Failed to create DataSource: {e}')
            return
        if verbose:
            print('Creating Features from DataSource')
        try:
            dimensions, features = dfutils.confirm_df_columns(df, dimensions, features)
            feature_set = self.features_from_source(source.id,
                                            features=features,
                                            dimensions=dimensions,
                                            granularity=granularity)
            features = self.get.features_by_featureset(feature_set.id)
            if verbose:
                print('Successfully created:', features)
        except Exception as e:
            print(f'Failed to create Features: {e}')
            return
        if verbose:
            print('Creating a Collection of your Features')
        try:
            collection = self.create.collection('name', 'Timeseries', granularity[0])
            collection.add_features(features)
            collection.generate_training_data()
            if verbose:
                print('Successfully created:', collection)
        except Exception as e:
            print(f'Failed to create Collection: {e}')
            return
        if not return_cli_only:
            webbrowser.open(f"app.rasgoml.com/collections/{collection.id}")
        return collection

    @track_usage
    def features(self, features_dict: dict,
                 trigger_stats: bool = True) -> FeatureSet:
        f"""
        Creates or updates Features based on metadata provided in the dict

        params:
            features_dict: Valid Rasgo dict (see below)

        return:
            Rasgo FeatureSet

        Valid Rasgo dict format:
        -----------------------
        {ingestion.RASGO_DICT}
        """
        # TODO need to publish docs on granularity and data types
        return self.save.feature_set_dict(features_dict, trigger_stats)

    @track_usage
    def features_from_df(self,
                         df: pd.DataFrame,
                         dimensions: List[str],
                         features: List[str],
                         granularity: List[str] = [],
                         tags: List[str] = [],
                         sandbox: bool = True):
        raise NotImplementedError("`publish.features_from_df()` is a deprecated method. "
                                  "Please use `publish.features()` or `publish.features_from_source()` instead")

    @track_usage
    def features_from_source(self,
                             data_source_id: int,
                             dimensions: List[str],
                             features: List[str] = [],
                             granularity: List[str] = [],
                             tags: List[str] = [],
                             feature_set_name: str = None,
                             sandbox: bool = True,
                             trigger_stats: bool = True,
                             if_exists: str = 'fail',
                             verbose: bool = False) -> FeatureSet:
        """
        Publishes a FeatureSet from an existing DataSource table

        params:
            data_source_id: ID to a Rasgo DataSource
            features: List of column names that will be features
                      If none is passed, all columns not in `dimensions` list will be registered as features
            dimensions: List of column names that will be dimensions
            granularity: List of grains that describe the form of the data. Common values are: 'day', 'week', 'month', 'second'
            tags: List of tags to be added to all features
            feature_set_name: Optional name for the FeatureSet (if not passed a random string will be assigned)
            sandbox: boolean: False = Production, True = Sandbox
            if_exists:  fail - returns an error message if a featureset already exists against this table
                        return - returns the featureset without operating on it
                        edit - edits the existing featureset
                        new - creates a new featureset
            verbose: boolean: Print status statements to stdout while function executes

        return:
            Rasgo FeatureSet
        """
        # V1 Trade-offs / Possible Future Enhancements
        # --------------
        # Constucting a featureset using v0 method - cut over to v1
        # We aren't adding display names, descriptions to features

        # Check for valid DataSource
        data_source = self.get.data_source(data_source_id)
        if not data_source.table:
            raise ValueError(f"DataSource {data_source_id} is not usable. Please make sure it exists and has a valid table registered.")
        if not isinstance(dimensions, list) and all([isinstance(dimension, str) for dimension in dimensions]):
            raise TypeError('Dimensions must be provided as a list of strings.')
        if not isinstance(features, list) and all([isinstance(feature, str) for feature in features]):
            raise TypeError('Features must be provided as a list of strings.')
        if not isinstance(granularity, list):
            if isinstance(granularity, str):
                granularity = [granularity]
            else:
                raise TypeError("granularity must be provided as a list of strings")
        if not isinstance(tags, list):
            if isinstance(tags, str):
                tags = [tags]
            else:
                raise TypeError("tags must be provided as a list of strings")
        if len(granularity) > len(dimensions):
            raise APIError("Number of granularities cannot exceed number of dimensions."
                           "Dimensions are the index fields your data should join on and group by."
                           "Granularity is an attribute that describes precision of your dimensions."
                           "Consider passing more dimensions or fewer granularities.")

        db = data_source.tableDatabase.upper() if data_source.tableDatabase else None
        schema = data_source.tableSchema.upper() if data_source.tableSchema else None
        table = data_source.table.split(".")[-1] if data_source.table.count(".") > 0 else data_source.table
        if verbose:
            print(f"Publishing features from source table {db}.{schema}.{table}")

        # Handle if FeatureSet already exists
        feature_set = self.match.feature_set(table_name=table)
        if feature_set and if_exists == 'fail':
            raise APIError(f"A featureset ({feature_set.id} - {feature_set.name}) is already built against this table. "
                            "Pass parameter if_exists = 'edit', 'new', or 'return' to continue.")

        columns = self.data_warehouse.get_source_columns(table=table, database=db, schema=schema)
        if columns.empty:
            raise APIError("The table associated with this DataSource is not accessible.")
        dimensions, features = dfutils.confirm_list_columns(columns["COLUMN_NAME"].values.tolist(), dimensions, features)

        # Publish Featureset V0
        timestamp = naming.make_timestamp()
        feature_set = self.save.feature_set(name=feature_set_name or data_source.name+"_"+timestamp,
                                            data_source_id=data_source.id,
                                            table_name=data_source.table,
                                            source_code=data_source.sourceCode,
                                            if_exists=if_exists)

        for _i, row in columns.iterrows():
            if row["COLUMN_NAME"] in dimensions:
                # Dimension / Granularity order matching:
                try:
                    # Try to match the position of the dim with the position of the granularity
                    dim_pos = dimensions.index(row['COLUMN_NAME'])
                    dim_granularity = granularity[dim_pos] if len(granularity) > 1 else granularity[0]
                except:
                    # Otherwise default to the next granularity in the list
                    dim_granularity = granularity.pop(0) if len(granularity) > 1 else granularity[0]
                self.save.dimension(feature_set_id=feature_set.id,
                                    column_name=row["COLUMN_NAME"],
                                    data_type=naming._cleanse_sql_data_type(row["DATA_TYPE"]),
                                    granularity=dim_granularity,
                                    if_exists=if_exists)

            if row["COLUMN_NAME"] in features:
                if verbose:
                    print(f'Publishing feature {row["COLUMN_NAME"]}')
                self.save.feature(feature_set_id=feature_set.id,
                                  display_name=row["COLUMN_NAME"],
                                  data_type=naming._cleanse_sql_data_type(row["DATA_TYPE"]),
                                  description=f"Feature {row['COLUMN_NAME']} created from DataSource {data_source.table}",
                                  column_name=row["COLUMN_NAME"],
                                  granularity=granularity[0],
                                  tags=tags,
                                  status="Sandboxed" if sandbox else "Productionized",
                                  if_exists=if_exists)

        if trigger_stats:
            if verbose:
                print("Triggering stats for all features. These may take a few minutes to run.")
            self.create.feature_set_stats(feature_set.id)
        return self.get.feature_set(feature_set.id)

    @track_usage
    def features_from_source_code(self,
                                  data_source_id: int,
                                  source_code_type: str,
                                  dimensions: List[str],
                                  features: List[str] = [],
                                  granularity: List[str] = [],
                                  tags: List[str] = [],
                                  sql_definition: str = None,
                                  python_function: FunctionType = None,
                                  feature_set_name: str = None,
                                  feature_set_table_name: str = None,
                                  sandbox: bool = True,
                                  sql_risk_override: bool = False,
                                  trigger_stats: bool = True,
                                  if_exists: str = 'fail',
                                  verbose: bool = False) -> FeatureSet:
        """
        Publishes a FeatureSet from an existing DataSource table

        params:
            data_source_id: ID to a Rasgo DataSource
            source_code_type: Valid string values are ['sql', 'python']
            sql_definition: Valid sql select statement that can create a view
            python_function: Valid python function that accepts a dataframe as input and returns a dataframe of features
            features: List of column names that will be features
                      If none is passed, all columns not in `dimensions` list will be registered as features
            dimensions: List of column names that will be dimensions
            granularity: List of grains that describe the form of the data. Common values are: 'day', 'week', 'month', 'second'
            tags: List of tags to be added to all features
            feature_set_name: Optional name for the FeatureSet (if not passed a random string will be assigned)
            sandbox: boolean: False = Production, True = Sandbox
            sql_risk_override: boolean: pass True if you receive a warning message about a dangerous sql keyword
            if_exists:  fail - returns an error message if a featureset already exists against this table
                        return - returns the featureset without operating on it
                        edit - edits the existing featureset
                        new - creates a new featureset
            verbose: boolean: Print status statements to stdout while function executes

        return:
            Rasgo FeatureSet
        """
        source_code_type = source_code_type.lower()
        if source_code_type not in ["sql", "python"]:
            raise ParameterValueError("source_code_type", ["sql", "python"])
        if source_code_type == "sql":
            if not sql_definition:
                raise ParameterValueError("sql_definition", ["Valid sql select statement that can create a veiw"])
        elif source_code_type == "python":
            raise APIError("Not supporting python functions quite yet...")
            if not python_function:
                raise ParameterValueError("python_function", ["Valid python function that accepts a dataframe as input and returns a dataframe of features"])

        # Publish the new data as a source
        # First make sure this is a valid source
        parent_source = self.get.data_source(data_source_id)
        if verbose:
            print(f"Publishing derivative source from parent source {parent_source.table}")
        child_source = self.source_data(source_type=source_code_type,
                                        sql_definition=sql_definition if source_code_type == "sql" else None,
                                        python_function=python_function if source_code_type == "python" else None,
                                        data_source_name=feature_set_name,
                                        data_source_table_name=feature_set_table_name,
                                        parent_data_source_id=data_source_id,
                                        sql_risk_override=sql_risk_override,
                                        trigger_stats=trigger_stats,
                                        if_exists=if_exists)

        # Then publish features
        if verbose:
            print(f"Publishing features from derivative source {child_source.id}")
        return self.features_from_source(data_source_id=child_source.id,
                                         features=features,
                                         dimensions=dimensions,
                                         granularity=granularity,
                                         tags=tags,
                                         feature_set_name=feature_set_name,
                                         sandbox=sandbox,
                                         trigger_stats=trigger_stats,
                                         if_exists=if_exists)

    @track_usage
    def features_from_yml(self,
                          yml_file: str,
                          sandbox: Optional[bool] = True,
                          git_repo: Optional[str] = None,
                          trigger_stats: bool = True,
                          verbose: bool = False) -> FeatureSet:
        """
        Publishes metadata about a FeatureSet to Pyrasgo

        params:
            yml_file: Rasgo compliant yml file that describes the featureset(s) being created
            sandbox: Status of the features (True = 'Sandboxed' | False = 'Productionized')
            git_repo: Filepath string to these feature recipes in git
            verbose: boolean: Print status statements to stdout while function executes

        return:
            Rasgo FeatureSet
        """
        with open(yml_file) as fobj:
            if verbose:
                print("Unpacking yml")
            contents = yaml.load(fobj, Loader=yaml.SafeLoader)
        if isinstance(contents, list):
            raise APIError("More than one feature set found, please pass in only one feature set per yml")
        else:
            if verbose:
                print("Publishing features")
            feature_set = self.save.feature_set_dict(contents, trigger_stats)
        return feature_set

    @track_usage
    def source_data(self,
                    source_type: str,
                    file_path: Path = None,
                    df: pd.DataFrame = None,
                    table: str = None,
                    table_database: Optional[str] = None,
                    table_schema: Optional[str] = None,
                    sql_definition: str = None,
                    python_function: FunctionType = None,
                    data_source_name: str = None,
                    data_source_domain: str = None,
                    data_source_table_name: str = None,
                    parent_data_source_id: Optional[int] = None,
                    sql_risk_override: bool = False,
                    trigger_stats: bool = True,
                    if_exists: str = 'fail',
                    verbose: bool = True
                    ) -> DataSource:
        """
        Push a csv, Dataframe, or table to a Snowflake table and register it as a Rasgo DataSource (TM)

        NOTES: csv files will import all columns as strings

        params:
            source_type: Values: ['csv', 'dataframe', 'table', 'sql', 'python']
            df: pandas DataFrame (only use when passing source_type = 'dataframe')
            file_path: full path to a file on your local machine (only use when passing source_type = 'csv')
            table: name of a valid Snowflake table in your Rasgo account (only use when passing source_type = 'table')
            table_database: Optional: name of the database of the table passed in 'table' param (only use when passing source_type = 'table')
            table_schema: Optional: name of the schema of the table passed in 'table' param (only use when passing source_type = 'table')
            sql_definition: Optional: valid SQL string that will create a new view in your DataWarehous
            python_function: Optional: valid python function that accepts a dataframe as input and returns a dataframe of features
            data_source_name: Optional name for the DataSource (if not provided a random string will be used)
            data_source_table_name: Optional name for the DataSource table in Snowflake (if not provided a random string will be used)
            data_source_domain: Optional domain for the DataSource (default is NULL)
            parent_data_source_id: Optional ID of a valid Rasgo DataSource that is a parent to this DataSource (default is NULL)
            sql_risk_override: boolean: pass True if you receive a warning message about a dangerous sql keyword
            if_exists: Values: ['fail', 'append', 'replace'] directs the function what to do if a DataSource already exists with this table name  (defaults to fail)
            verbose: boolean: Print status statements to stdout while function executes

        return:
            Rasgo DataSource
        """
        # V1 Trade-offs / Possible Future Enhancements
        # --------------
        # csv's upload with all columns as string data type
        # uploading csv locally vs. calling the post/data-source/csv endpoint

        # Validate inputs
        vals = ["csv", "dataframe", "table", "sql"]
        source_type = source_type.lower()
        _org_defaults = self.api._get_default_namespace()

        if source_type not in vals:
            raise ParameterValueError("sourceType", vals)
        if source_type == "csv" and not Path(file_path).exists():
                raise FileNotFoundError("Please pass in a valid file path using the `file_path` parameter")
        if source_type == "dataframe":
            if not isinstance(df, pd.DataFrame) or df.empty:
                raise ValueError("Please pass in a valid DataFrame using the `df` parameter")
        if source_type == 'sql':
            if not sql_definition:
                raise ValueError('Please pass in a valid SQL string using the `sql_definition` parameter')
            #check for dangerous sql
            if naming._is_scary_sql(sql_definition) and not sql_risk_override:
                # NOTE: In the future we will want to send an email to the account owner
                #       or require special permissions to run with a sql_risk_override param
                raise ValueError("It looks like the sql string you passed may have some potentially dangerous keywords in it like DELETE or DROP or TRUNCATE. "
                                 "This SQL will be used to create a view in your DataWarehouse. Please ensure this is the code you want to run. "
                                 "If you're confident in your code, please pass in the parameter `sql_risk_override=True` to run this code.")
            if not naming._is_valid_view_sql(sql_definition) and not sql_risk_override:
                raise ValueError("It looks like the sql string you passed is missing basic elements like a SELECT or FROM clause. "
                                 "This SQL will be used to create a view in your DataWarehouse. Please ensure this is the code you want to run. "
                                 "If you're confident in your code, please pass in the parameter `sql_risk_override=True` to run this code.")
        if source_type == "table":
            if not table:
                raise ValueError("Please pass in a valid table using the `table` parameter")
            if if_exists in ["append", "replace"]:
                # Raise this as an error after we've confirmed there will be no backwards compat issues
                print("Data Sources built from SQL tables do not physically append or replace data, proceeding with a no-op registration.")
            parsed_database, parsed_schema, parsed_table = naming.parse_fqtn(table, _org_defaults)
            table = parsed_table
            table_database = table_database or parsed_database
            table_schema = table_schema or parsed_schema
            try:
                src_table = self.data_warehouse.get_source_table(table_name=table, database=table_database, schema=table_schema, record_limit=10)
                if src_table.empty:
                    raise APIError(f"Source table {table} is empty or this role does not have access to it.")
            except:
                raise APIError(f"Source table {table} does not exist or this role does not have access to it.")
        if data_source_table_name == "":
            raise ValueError("The `data_source_table_name` parameter was passed as a blank string, please pass in a valid SQL table name")

        # Determine the ouput table based on user inputs
        if source_type in ["csv", "dataframe", "sql"]:
            table_name = data_source_table_name or naming.random_table_name()
        else: #source_type == "table":
            table_name = table
        table_name = table_name.upper()
        table_database = table_database if source_type == "table" else _org_defaults.get('database')
        table_schema = table_schema if source_type == "table" else _org_defaults.get('schema')
        table_fqtn = naming.make_fqtn(table=table_name, database=table_database, schema=table_schema)

        # Check if a DataSource already exists
        data_source = self.match.data_source(fqtn=table_fqtn)
        if data_source:
            # If it does, override to the existing table
            table_name = data_source.table
            table_database = data_source.tableDatabase
            table_schema = data_source.tableSchema
            table_fqtn = naming.make_fqtn(table=table_name, database=table_database, schema=table_schema)

            # Then handle input directives
            vals = ["append", "fail", "replace"]
            msg = f"DataSource {data_source.id} already exists. "
            if data_source.organizationId != self.api._profile.get('organizationId'):
                raise APIError(msg+"This API key does not have permission to replace it.")
            if data_source.sourceType != source_type:
                raise APIError(msg+"Your input parameters would edit the source_type. To change DataSource attributes, use the update_data_source method. To update this source table, ensure your input parameters match the DataSource definition.")
            if if_exists not in vals:
                raise ParameterValueError("if_exists", vals)
            if if_exists == 'fail':
                raise APIError(msg+"Pass if_exists='replace' or 'append' to proceed.")

        # Determine operation to perform: [create, append, replace, register, no op]
        _operation = self._source_table_operation(source_type, if_exists, table_fqtn, org_defaults=_org_defaults)
        if verbose:
            print(f"Proceeding with operation {_operation}")

        # Upload to Snowflake
        # Path: csv & dataframe
        if source_type in ["csv", "dataframe"]:
            if source_type == "csv":
                df = pd.read_csv(file_path)
            if verbose:
                print("Uploading to table:", table_name)
            dfutils._cleanse_sql_dataframe(df)
            self.data_warehouse.write_dataframe_to_table(df, table_name=table_name, append=True if if_exists=="append" else False)
            self.data_warehouse.grant_table_ownership(table=table_name, role=self.data_warehouse.publisher_role)
            self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role)

        # Path: python
        if source_type == "python":
            raise NotImplementedError("Not supporting python source code just yet...")

        # Path: sql
        if source_type == "sql":
            view_query = f"CREATE OR REPLACE VIEW {table_name} AS {sql_definition};"
            if verbose:
                print("Creating sql view as:", view_query)
            self.data_warehouse.execute_query(view_query)
            self.data_warehouse.grant_table_ownership(table=table_name, role=self.data_warehouse.publisher_role)
            self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role)

        # Path: table
        if source_type == "table":
            if verbose:
                print(f"Granting read access on table {table_name} to {self.data_warehouse.reader_role.upper()}")
            # NOTE: grant_table_acess is intentional here
            # In other methods, we create a table with the rasgo user role and want to hand if off to the reader role
            # In this case, the table is likely part of a pre-existing rbac model and we just want to grant rasgo access
            self.data_warehouse.grant_table_access(table=table_name, role=self.data_warehouse.reader_role, database=table_database, schema=table_schema)

        # Publish DataSource
        data_source = self.save.data_source(name=data_source_name or table_name,
                                            table=table_name,
                                            database=table_database if source_type == 'table' else _org_defaults.get('database'),
                                            schema=table_schema if source_type == 'table' else _org_defaults.get('schema'),
                                            source_code=sql_definition if source_type == 'sql' else None,
                                            domain=data_source_domain,
                                            source_type=source_type,
                                            parent_source_id=parent_data_source_id,
                                            if_exists='edit')
        if data_source:
            if trigger_stats:
                self.create.data_source_stats(data_source.id)
            return data_source
        else:
            raise APIError("DataSource failed to upload")

    def _source_table_operation(self,
                                source_type: str,
                                if_exists: str,
                                to_fqtn: str,
                                from_fqtn: Optional[str] = None,
                                org_defaults: Optional[dict] = None):
        """
        Called by publish_source_data:
            Given a source_type and tables, return the operation that should be performed to publish this table
        """
        to_database, to_schema, to_table = naming.parse_fqtn(to_fqtn, org_defaults)
        data_source_exists = True if self.match.data_source(fqtn=to_fqtn) is not None else False

        try:
            dest_table = self.data_warehouse.get_source_table(table_name=to_table, database=to_database, schema=to_schema, record_limit=10)
            is_dest_table_empty = dest_table.empty
        except:
            is_dest_table_empty = True

        if source_type in ["csv", "dataframe", "sql"]:
            if not data_source_exists:
                if is_dest_table_empty:
                    return "create"
                else: #not is_dest_table_empty
                    raise APIError(f"A table named {to_fqtn} already exists, but is not registered as a Rasgo DataSource. "
                                   f"Try running this function with params: source_type='table', table='{to_table}'. "
                                    "If this wasn't an intentional match, run this function again to generate a new table name.")
            elif data_source_exists:
                if is_dest_table_empty:
                    return "create"
                else: #not is_dest_table_empty
                    return if_exists
            else:
                raise APIError("Could not determine what operation to perform.")
        elif source_type == "table":
            if not data_source_exists:
                return "register"
            elif not is_dest_table_empty and if_exists in ["append", "replace"]:
                print(f"pyRasgo does not support {if_exists} operations on tables yet.")
            return "no op"
        else:
            raise APIError("Could not determine what operation to perform.")


import pandas as pd
from typing import Dict, List, Optional

from .error import APIError
from pyrasgo.primitives import Feature, FeatureList, FeatureSet, DataSource
from pyrasgo import schemas as api
from pyrasgo.utils import ingestion, naming, track_usage


class Save():

    def __init__(self):
        from . import Get, Create, Match, Update
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()
        self.get = Get()
        self.match = Match()
        self.create = Create()
        self.update = Update()

    @track_usage
    def data_source(self, table: str,
                    name: str = None,
                    database: Optional[str] = None,
                    schema: Optional[str] = None,
                    source_code: Optional[str] = None,
                    domain: Optional[str] = None,
                    source_type: Optional[str] = None,
                    parent_source_id: Optional[int] = None,
                    if_exists: str = 'return') -> DataSource:
        """
        Creates or returns a DataSource depending on of the defined parameters

        if_exists:  return - returns the source without operating on it
                    edit - edits the existing source
                    new - creates a new source
        """
        fqtn = naming.make_fqtn(table=table, database=database, schema=schema)
        ds = self.match.data_source(fqtn=fqtn)
        if ds:
            if if_exists == 'return':
                return ds
            elif if_exists in ['edit', 'replace', 'append']:
                return self.update.data_source(id=ds.id, name=name, table=table, database=database, schema=schema, source_code=source_code, domain=domain, source_type=source_type, parent_source_id=parent_source_id)
            else:
                raise APIError(f"Table {fqtn} is already registered as Data Source {ds.id}")
        return self.create.data_source(name=name, table=table, database=database, schema=schema, source_code=source_code, domain=domain, source_type=source_type, parent_source_id=parent_source_id)

    @track_usage
    def dataframe(self, unique_id: str,
                  name: str = None,
                  shared_status: str = None,
                  column_hash: Optional[str] = None,
                  update_date: str = None,
                  if_exists: str = 'return') -> api.Dataframe:
        """
        Creates or returns a Dataframe depending on of the defined parameters

        if_exists:  return - returns the source without operating on it
                    edit - edits the existing source
                    new - creates a new source
        """
        df = self.match.dataframe(unique_id=unique_id)
        if df:
            if if_exists == 'return':
                return df
            elif if_exists in ['edit', 'replace', 'append']:
                return self.update.dataframe(df.uniqueId, name, shared_status, column_hash, update_date)
            else:
                raise APIError(f"ID {unique_id} is already registered as DataFrame {df.id}")
        return self.create.dataframe(unique_id, name, shared_status, column_hash, update_date)

    @track_usage
    def dimension(self, feature_set_id: int,
                  column_name: str,
                  data_type: str,
                  display_name: Optional[str] = None,
                  dimension_type: Optional[str] = None,
                  granularity: Optional[str] = None,
                  if_exists: str = 'return') ->api.Column:
        """
        Creates or updates a dimension depending on existence of the defined parameters

        if_exists:  return - returns the source without operating on it
                    edit - edits the existing source
                    new - creates a new source
        """
        dimensionality = self.dimensionality(dimension_type=dimension_type, granularity=granularity)
        dim = self.match.column(name=column_name, feature_set_id=feature_set_id)
        if dim:
            if if_exists == 'return':
                return dim
            elif if_exists in ['edit', 'replace', 'append']:
                return self.update.column(id=dim.id, name=column_name, data_type=data_type, feature_set_id=feature_set_id, dimension_id=dimensionality.id)
            else:
                raise APIError(f"Column {column_name} is already registered in FeatureSet {feature_set_id} as Dimension {dim.id}")
        return self.create.column(name=column_name, data_type=data_type, feature_set_id=feature_set_id, dimension_id=dimensionality.id)

    @track_usage
    def dimensionality(self, granularity: str,
                       dimension_type: Optional[str] = None,
                       if_exists: str = 'return') -> api.Dimensionality:
        """
        Creates or returns a dimensionality depending on existence of the defined parameters

        Dimensionality is a named pairing of a datatype and a granularity. Note in some cases the
        granularity is actually a data type.
        """
        # TODO: We should move this mapping to the Granularity enum class or behind an API
        if dimension_type is None:
            if granularity.lower() in ["second", "minute", "hour", "day", "week", "month", "quarter", "year"]:
                dimension_type = "DateTime"
            elif granularity.lower() in ["latlong", "zipcode", "fips", "dma", "city", "cbg", "county", "state",
                                         "country"]:
                dimension_type = "Geolocation"
            else:
                dimension_type = "Custom"
        elif dimension_type.lower() == "datetime":
            dimension_type = "DateTime"
        elif dimension_type.lower() in ["geo", "geoloc", "geolocation"]:
            dimension_type = "Geolocation"
        else:
            dimension_type = dimension_type.title()

        dimensionality = self.match.dimensionality(granularity)
        # No edit path for dimensionality for now...
        if dimensionality:
            if if_exists in ['return', 'edit']:
                return dimensionality
            else:
                raise APIError(f"Dimensionality {granularity} is already registered as Dimensionality {dimensionality.id}")
        return self.create.dimensionality(dimension_type=dimension_type, granularity=granularity)

    @track_usage
    def feature(self, feature_set_id: int,
                column_name: str,
                display_name: str,
                data_type: str,
                description: Optional[str] = None,
                granularity: Optional[str] = None,
                status: Optional[str] = None,
                tags: Optional[List[str]] = None,
                git_repo: Optional[str] = None,
                if_exists: str = 'return') -> Feature:
        """
        Creates or updates a feature depending on existence of the defined parameters

        if_exists:  return - returns the source without operating on it
                    edit - edits the existing source
                    new - creates a new source
        """
        column_name = column_name or display_name
        description = description or f"Feature that contains {display_name} data"
        status = status or "Sandboxed"
        dimension_id = None if granularity is None else self.dimensionality(dimension_type=None, granularity=granularity).id

        ft = self.match.feature(code=column_name, feature_set_id=feature_set_id)
        if ft:
            if if_exists == 'return':
                return ft
            elif if_exists in ['edit', 'replace', 'append']:
                self.update.column(id=ft.columnId, name=column_name, data_type=data_type, feature_set_id=feature_set_id, dimension_id=dimension_id)
                return self.update.feature(id=ft.id, display_name=display_name, column_name=column_name, description=description, status=status, tags=tags or [], git_repo=git_repo)
            else:
                raise APIError(f"Column {column_name} is already registered in FeatureSet {feature_set_id} as Feature {ft.id}")
        column = self.create.column(name=column_name, data_type=data_type, feature_set_id=feature_set_id, dimension_id=dimension_id)
        return self.create.feature(feature_set_id=feature_set_id, display_name=display_name, column_name=column_name, description=description, column_id=column.id, status=status, git_repo=git_repo, tags=tags or [])

    @track_usage
    def feature_set(self, data_source_id: int,
                    table_name: str,
                    name: str,
                    file_path: Optional[str] = None,
                    source_code: Optional[str] = None,
                    if_exists: str = 'return') -> FeatureSet:
        """
        Creates or updates a featureset depending on existence of the defined parameters

        if_exists:  return - returns the source without operating on it
                    edit - edits the existing source
                    new - creates a new source
        """
        fs = self.match.feature_set(table_name=table_name)
        if fs:
            if if_exists == 'return':
                return fs
            elif if_exists in ['edit', 'replace', 'append']:
                return self.update.feature_set(id=fs.id, name=name, data_source_id=data_source_id, table_name=table_name, file_path=file_path, source_code=source_code)
            else:
                raise APIError(f"Table {table_name} is already registered as FeatureSet {fs.id}")
        return self.create.feature_set(name=name, data_source_id=data_source_id, table_name=table_name, file_path=file_path, source_code=source_code)

    @track_usage
    def feature_set_dict(self, feature_set_dict: dict,
                         trigger_stats: bool = True) -> FeatureSet:
        """
        Creates or updates a featureset based on values in a dict
        """
        if not ingestion._confirm_valid_dict(dict_in=feature_set_dict):
            raise APIError("Not a valid dict")
        source_table = feature_set_dict.get("sourceTable")
        ds = feature_set_dict.get("datasource")
        ds_name = ds.get("name", source_table) if ds else source_table
        data_source = self.data_source(name=ds_name, table=source_table, source_type="Table")
        featureset_name = feature_set_dict.get("name", source_table)
        tags = []
        if feature_set_dict.get("tags"):
            for t in feature_set_dict.get("tags"):
                tags.append(t)
        attributes = []
        if feature_set_dict.get("attributes"):
            for a in feature_set_dict.get("attributes"):
                for k, v in a.items():
                    attributes.append({k: v})
        source_code = feature_set_dict.get("sourceCode")
        featureset = self.feature_set(name=featureset_name, data_source_id=data_source.id, table_name=source_table, source_code=source_code, if_exists='edit')

        # publish dimensions
        for dim in feature_set_dict["dimensions"]:
            column_name = dim.get("columnName")
            display_name = dim.get("displayName")
            data_type = dim.get("dataType")
            # if we get an enum, convert it to str so pydantic doesn't get mad
            if isinstance(data_type, api.DataType):
                data_type = data_type.value
            dim_granularity = dim.get("granularity")
            self.dimension(feature_set_id=featureset.id, column_name=column_name, display_name=display_name, data_type=data_type, granularity=dim_granularity, if_exists='edit')

        # publish features
        for feature in feature_set_dict["features"]:
            display_name = feature.get("displayName")
            column_name = feature.get("columnName")
            data_type = feature.get("dataType")
            # if we get an enum, convert it to str so pydantic doesn't get mad
            if isinstance(data_type, api.DataType):
                data_type = data_type.value
            description = feature.get("description", f"Feature that contains {display_name} data")
            # apply featureset tags to all features...
            feature_tags = tags
            # ...and add feature-specific tags
            if feature.get("tags"):
                for t in feature.get("tags"):
                    feature_tags.append(t)
            feature_attributes = attributes
            if feature.get("attributes"):
                for a in feature.get("attributes"):
                    for k, v in a.items():
                        feature_attributes.append({k: v})
            status = "Sandboxed" if feature_set_dict.get("status") == "Sandboxed" else "Productionized"
            f = self.feature(feature_set_id=featureset.id, display_name=display_name, data_type=data_type, column_name=column_name, description=description, status=status, tags=feature_tags, if_exists='edit')
            self.update.feature_attributes(f.id, feature_attributes)

        # Post stats for features
        if trigger_stats:
            self.create.feature_set_stats(featureset.id)
        return self.get.feature_set(featureset.id)

    @track_usage
    def column_importance_stats(self, id: str, payload: api.feature.featureImportanceStats):
        """
        Sends a json payload of importance from a dataFrame to the API so it can render in the WebApp
        """
        # First update timestamp on the DF object
        self.dataframe(unique_id=id, update_date=payload.timestamp, if_exists='edit')
        return self.create.column_importance_stats(id, payload)

    @track_usage
    def feature_importance_stats(self, id: int, payload: api.feature.featureImportanceStats):
        """
        sends a JSON payload of importance from a collection to the API so it can store and render stats later
        """
        return self.create.feature_importance_stats(id=id, payload=payload)

    @track_usage
    def dataframe_profile(self, id: str, payload: api.feature.ColumnProfiles):
        """
        Send a json payload of a dataframe profile so it can render in the WebApp
        """
        # First update timestamp on the DF object
        self.dataframe(unique_id=id, update_date=payload.timestamp, if_exists='edit')
        return self.create.dataframe_profile(id, payload)


from typing import Optional

from pyrasgo import config
from pyrasgo.api.create import Create
from pyrasgo.api.match import Match
from pyrasgo.api.publish import Publish
from pyrasgo.utils import track_usage

class RasgoOrchestration():

    def __init__(self):
        self.match = Match()
        self.create = Create()
        self.publish = Publish()

    @track_usage
    def get_featureset_id(self, snowflakeTable: str, fs_name: Optional[str] = None) -> Optional[int]:
            fs = self.match.feature_set(table_name=snowflakeTable)
            return fs.id if fs else None

    @track_usage
    def publish_features_from_yml(self, yml_file: str, sandbox: Optional[bool] = True, git_repo: Optional[str] = None):
        return self.publish.features_from_yml(yml_file=yml_file, sandbox=sandbox, git_repo=git_repo)

    @track_usage
    def run_stats_for_feature(self, feature_id: int):
        return self.create.feature_stats(feature_id)

    @track_usage
    def run_stats_for_featureset(self, featureset_id: int):
        return self.create.feature_set_stats(featureset_id)


import pandas as pd
from typing import Dict, List, Optional

from .error import APIError
from pyrasgo import config
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, DataSource
from pyrasgo.utils import track_usage


class Read():
    
    def __init__(self):
        from . import Get
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse
        
        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()
        self.get = Get()

    @track_usage
    def collection_data(self, 
                        id: int,
                        filters: Optional[Dict[str, str]] = None,
                        limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Collection

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        collection = self.get.collection(id)
        if collection:
            try:
                table_metadata = collection._make_table_metadata()
                query, values = self.data_warehouse._make_select_statement(table_metadata, filters, limit)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"Collection table is not reachable: {e}")
        raise APIError("Collection does not exist")

    @track_usage
    def feature_data(self, 
                     id: int,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Feature data

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        feature = self.get.feature(id)
        if feature.sourceTable:
            try:
                table_metadata = feature._make_table_metadata()
                #TODO: if we ever support multiple features, add them to this line -
                features = feature.columnName
                indices = ','.join(feature.indexFields)
                columns = indices +', '+features
                query, values = self.data_warehouse._make_select_statement(table_metadata, filters, limit, columns)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"Feature table is not reachable: {e}")
        raise APIError("Feature table does not exist")

    @track_usage
    def source_data(self, 
                    id: int,
                    filters: Optional[Dict[str, str]] = None,
                    limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo DataSource

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        data_source = self.get.data_source(id)
        if data_source:
            try:
                table_metadata = data_source._make_table_metadata()
                query, values = self.data_warehouse._make_select_statement(table_metadata, filters, limit)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"DataSource table is not reachable: {e}")
        raise APIError("DataSource does not exist")


from pyrasgo.api.connection import Connection
from pyrasgo.schemas import feature_set as api
from pyrasgo.utils import track_usage

class FeatureSet(Connection):
    """
    Stores a Rasgo FeatureSet
    """

    def __init__(self, api_object, **kwargs):
        super().__init__(**kwargs)
        self._fields = api.FeatureSet(**api_object)

    def __getattr__(self, item):
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            self.refresh()
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"FeatureSet(id={self.id}, name={self.name}, sourceTable={self.sourceTable}, granularities={self.granularities})"

# ----------
# Properties
# ----------

# -------
# Methods
# -------
    @track_usage
    def display_source_code(self):
        """
        Convenience function to display the sourceCode property
        """
        return self._fields.sourceCode

    @track_usage
    def rebuild_from_source_code(self):
        """
        Rebuild the FeatureSet using the source code
        """
        raise NotImplementedError()

    @track_usage
    def refresh(self):
        """
        Updates the FeatureSet's attributes from the API
        """
        self._fields = api.FeatureSet(**self._get(f"/feature-sets/{self.id}", api_version=1).json())

    @track_usage
    def rename(self, new_name: str):
        """
        Updates a FeatureSet's display name
        """
        print(f"Renaming FeatureSet {self.id} from {self.name} to {new_name}")
        featureset = api.FeatureSetUpdate(id=self.id, name=new_name)
        self._fields = api.FeatureSet(**self._patch(f"/feature-sets/{self.id}",
                                                    api_version=1, _json=featureset.dict(exclude_unset=True, exclude_none=True)).json())

    def _make_table_metadata(self):
        table_attribute = self._fields.sourceTable
        organization = self._get_profile().get("organization")
        if table_attribute.count(".") == 2:
            database = table_attribute.split(".")[0]
            schema = table_attribute.split(".")[1]
            table = table_attribute.split(".")[-1]
        else:
            database = organization.get("database")
            schema = organization.get("schema")
            table = table_attribute
        metadata = {
            "database": database,
            "schema": schema,
            "table": table,
        }
        return metadata


from pydantic import BaseModel, Field
from typing import List, Optional

from pyrasgo.schemas.column import Column
from pyrasgo.schemas.enums import ModelTypes
from pyrasgo.schemas.feature import Feature


class CollectionBase(BaseModel):
    id: int

class Collection(CollectionBase):
    name: Optional[str]
    authorId: Optional[int]
    dataTableName: Optional[str]
    type: Optional[ModelTypes] = None
    organizationId: Optional[int]
    dimensions: Optional[List[Column]]
    features: Optional[List[Feature]]
    description: Optional[str]
    isShared: Optional[bool]

class CollectionUpdate(BaseModel):
    id: int
    name: Optional[str]


from typing import Optional
from pydantic import BaseModel
from pyrasgo.schemas.dimensionality import Dimensionality


class ColumnCreate(BaseModel):
    name: str
    dataType: str
    featureSetId: int
    dimensionId: Optional[int]


class ColumnUpdate(BaseModel):
    id: int
    name: Optional[str]
    dataType: Optional[str]
    featureSetId: Optional[int]
    dimensionId: Optional[int]


class Column(BaseModel):
    id: Optional[int]
    name: Optional[str]
    dataType: Optional[str]
    featureSetId: Optional[int]
    dimensionId: Optional[int]
    granularity: Optional[str]